import click
import datetime
import pdb
import os
import re
import requests
import sys

from urllib.parse import urlparse
from typing import List

from stoobly_agent.app.models.adapters.raw_joined import RawJoinedRequestAdapterFactory
from stoobly_agent.app.models.factories.resource.local_db.helpers.log import Log
from stoobly_agent.app.models.factories.resource.local_db.helpers.log_event import LogEvent, PUT_ACTION, REQUEST_RESOURCE, SCENARIO_RESOURCE
from stoobly_agent.app.models.factories.resource.local_db.helpers.request_snapshot import RequestSnapshot
from stoobly_agent.app.models.factories.resource.local_db.helpers.scenario_snapshot import ScenarioSnapshot
from stoobly_agent.app.models.helpers.apply import Apply
from stoobly_agent.config.data_dir import DataDir
from stoobly_agent.lib.api.keys import RequestKey, ScenarioKey

from .helpers.print_service import FORMATS, print_snapshots, select_print_options
from .helpers.verify_raw_request_service import verify_raw_request

@click.group(
    epilog="Run 'stoobly-agent project COMMAND --help' for more information on a command.",
    help="Manage snapshots."
)
@click.pass_context
def snapshot(ctx):
    pass

@snapshot.command(
  help="Apply snapshots.",
)
@click.option('--force', is_flag=True, default=False, help="Toggles whether resources are hard deleted.")
@click.argument('uuid', required=False)
def apply(**kwargs):
  apply = Apply(force=kwargs['force']).with_logger(print)
  completed = True

  if kwargs.get('uuid'):
    completed = apply.single(kwargs['uuid'])
  else:
    completed = apply.all()

  if not completed:
    sys.exit(1)

@snapshot.command(
  help="Copy snapshots to a different data directory."
)
@click.option('--request-key', multiple=True, help='')
@click.option('--scenario-key', multiple=True, help='')
@click.argument('destination', required=True)
def copy(**kwargs):
  destination = kwargs['destination']
  __copy_scenarios(kwargs['scenario_key'], destination)
  __copy_requests(kwargs['request_key'], destination)

@snapshot.command(
  help="List snapshots.",
  name="list"
)
@click.option('--format', type=click.Choice(FORMATS), help='Format output.')
@click.option('--pending', default=False, is_flag=True, help='Lists not yet processed snapshots.')
@click.option(
  '--resource',
  default=REQUEST_RESOURCE,
  type=click.Choice([REQUEST_RESOURCE, SCENARIO_RESOURCE]),
  help=f"Defaults to {REQUEST_RESOURCE}."
)
@click.option('--scenario', help='Scenario name regex pattern to filter snapshots by')
@click.option('--search', help='Regex pattern to filter snapshots by.')
@click.option('--select', multiple=True, help='Select column(s) to display.')
@click.option('--size', default=10, help='Number of rows to display.')
@click.option('--without-headers', is_flag=True, default=False, help='Disable printing column headers.')
def _list(**kwargs):
  print_options = select_print_options(kwargs)

  log = Log()

  events = None
  if kwargs.get('pending'):
    events = log.unprocessed_events
  else:
    events = log.target_events

  if events:
    formatted_events = __format_events(events, **kwargs)

    if len(formatted_events):
      print_snapshots(formatted_events, **print_options)

@snapshot.command(
  help="Prune deleted snapshots."
)
@click.option('--dry-run', is_flag=True, default=False)
def prune(**kwargs):
  log = Log()
  log.prune(kwargs['dry_run'])

@snapshot.command(
  help="Update a snapshot.",
)
@click.option('--format', type=click.Choice(FORMATS), help='Format output.')
@click.option('--select', multiple=True, help='Select column(s) to display.')
@click.option('--verify', is_flag=True, default=False)
@click.option('--without-headers', is_flag=True, default=False, help='Disable printing column headers.')
@click.argument('uuid')
def update(**kwargs):
  print_options = select_print_options(kwargs)

  log = Log()

  event = None
  for _event in log.events:
    if _event.uuid == kwargs['uuid']:
      event = _event
      break

  if not event:
    print(f"Error: {kwargs['uuid']} not found", file=sys.stderr)
    sys.exit(1)

  if kwargs['verify']:
    if event.is_request(): 
      snapshot: RequestSnapshot = event.snapshot()
      __verify_request(snapshot)
    elif event.is_scenario():
      snapshot: ScenarioSnapshot = event.snapshot()
      snapshot.iter_request_snapshots(__verify_request)

  new_event = event.duplicate()
  log.append(str(new_event))

  formatted_events = __format_events([new_event], **kwargs)

  if len(formatted_events):
    print_snapshots(formatted_events, **print_options)

def __format_events(events: List[LogEvent], **kwargs):
  count = 0
  formatted_events = []
  size = 10 if kwargs.get('size') == None else kwargs['size']

  if kwargs.get('resource') == SCENARIO_RESOURCE:
    for event in events:
      if count == size:
        break

      if event.resource != SCENARIO_RESOURCE:
        continue

      if not __scenario_matches(event, kwargs.get('scenario')):
        continue

      snapshot = event.snapshot()
      if not __description_matches(snapshot, kwargs.get('search')):
        continue

      path = os.path.relpath(snapshot.metadata_path)

      formatted_events.append({
        **__transform_scenario(snapshot),
        'snapshot': path,
        **__transform_event(event)
      })

      count += 1
  else:
    joined_events = []
    for event in events:
      if event.resource != REQUEST_RESOURCE:
        snapshot: ScenarioSnapshot = event.snapshot()
        snapshot.iter_request_snapshots(lambda snapshot: joined_events.append((event, snapshot))) 
      else:
        joined_events.append((event, event.snapshot()))

    for joined_event in joined_events:
      if count == size:
        break

      event, snapshot = joined_event
      request = __to_request(snapshot)

      if not __scenario_matches(event, kwargs.get('scenario')):
        continue

      if not __request_matches(request, kwargs.get('search')):
        continue

      path = os.path.relpath(snapshot.path)
      scenario = ''

      if event.resource == SCENARIO_RESOURCE:
        snapshot: ScenarioSnapshot = event.snapshot()
        scenario = snapshot.metadata.get('name')

      formatted_events.append({
        **__transform_request(request),
        'snapshot': path,
        **__transform_event(event),
        'scenario': scenario
      })

      count += 1

  return formatted_events

def __verify_request(snapshot: RequestSnapshot):
  raw_request = snapshot.request
  if not raw_request:
    return

  verified_raw_request = verify_raw_request(raw_request)

  if raw_request != verified_raw_request:
    snapshot.write_raw(verified_raw_request)

def __transform_event(event: LogEvent):
  event_dict = {}

  event_dict['uuid'] = event.uuid
  event_dict['action'] = event.action

  if event.created_at:
    event_dict['created_at'] = datetime.datetime.fromtimestamp(event.created_at / 1000)

  return event_dict

def __to_request(snapshot: RequestSnapshot):
  raw_request = snapshot.request
  if not raw_request:
    return None

  return RawJoinedRequestAdapterFactory(raw_request).python_request()

def __request_matches(request: requests.Request, search: str):
  if not search:
    return True
  
  if not request:
    return False

  uri = urlparse(request.url)
  return re.match(search, request.url) or re.match(search, uri.path)

def __transform_request(request: requests.Request):
  event_dict = { 'method': '', 'host': '', 'port': '', 'path': '', 'query': ''}

  if request:
    uri = urlparse(request.url)
    event_dict['method'] = request.method
    event_dict['host'] = uri.hostname
    event_dict['port'] = uri.port
    event_dict['path'] = uri.path
    event_dict['query'] = uri.query

  return event_dict

def __description_matches(snapshot: ScenarioSnapshot, search: str):
  if not search:
    return True

  metadata = snapshot.metadata
  return re.match(search, metadata.get('description') or '')

def __scenario_matches(event: LogEvent, search: str):
  if not search:
    return True

  if event.resource != SCENARIO_RESOURCE:
    return False

  snapshot = event.snapshot()
  metadata = snapshot.metadata
  return re.match(search, metadata.get('name') or '')

def __transform_scenario(snapshot: ScenarioSnapshot):
  event_dict = {}

  metadata = snapshot.metadata
  event_dict['name'] = metadata.get('name')
  event_dict['description'] = metadata.get('description')

  return event_dict

def __copy_requests(request_keys: list, destination: str):
  log = Log()

  data_dir = DataDir.instance(destination)
  destination_log = Log(data_dir)

  for request_key in request_keys:
    found = False

    for event in log.target_events:
      if event.action != PUT_ACTION:
        continue

      if event.resource != REQUEST_RESOURCE:
        continue

      key = RequestKey(request_key)
      if event.resource_uuid != key.id:
        continue

      snapshot: RequestSnapshot = event.snapshot()
      snapshot.copy(destination)
      resource = snapshot.find_resource()

      if not resource:
        print(f"Could not find request {key.id}", file=sys.stderr)
      else:
        destination_log.put(resource)
        found = True

      if not found:
        print(f"No snapshot found for {key}", file=sys.stderr)

def __copy_scenarios(scenario_keys: list, destination: str):
  log = Log()

  data_dir =  DataDir.instance(destination)
  destination_log = Log(data_dir)

  for scenario_key in scenario_keys:
    found = False

    for event in log.target_events:
      if event.action != PUT_ACTION:
        continue

      if event.resource != SCENARIO_RESOURCE:
        continue

      key = ScenarioKey(scenario_key)
      if event.resource_uuid != key.id:
        continue

      snapshot: ScenarioSnapshot = event.snapshot()
      snapshot.copy(destination)
      resource = snapshot.find_resource()

      if not resource:
        print(f"Could not find scenario {key.id}", file=sys.stderr)
      else:
        destination_log.put(resource)
        found = True

      if not found:
        print(f"No snapshot found for {key}", file=sys.stderr)


import pdb

from mitmproxy.coretypes.multidict import MultiDict
from mitmproxy.http import Headers, Request as MitmproxyRequest

from stoobly_agent.app.models.schemas.request import Request
from stoobly_agent.lib.utils.decode import decode

class MitmproxyRequestAdapter():

  def __init__(self, request: Request):
    self.__http_version = request.http_version
    self.__request = request

  @property
  def headers(self):
    _headers = self.__request.headers
    return Headers(**self.__decode_dict(_headers))

  def adapt(self):
    request = MitmproxyRequest.make(
      self.__request.method,
      self.__request.url,
      self.__request.body,
      self.headers,
    )
    request.http_version = self.__http_version
    return request

  def __decode_dict(self, d: Headers) -> MultiDict:
    new_d = MultiDict()
    for k, v in d.items():
      new_d.add(decode(k), decode(v))
    return new_d


import pdb

from mitmproxy.http import Request as MitmproxyRequest
from typing import TypedDict

from stoobly_agent.app.proxy.intercept_settings import InterceptSettings
from stoobly_agent.app.proxy.replay.context import ReplayContext
from stoobly_agent.config.constants import lifecycle_hooks, replay_policy

from .utils.allowed_request_service import get_active_mode_policy
from .utils.rewrite import rewrite_request, rewrite_response

LOG_ID = 'HandleReplay'

class ReplayOptions(TypedDict):
    no_rewrite: bool

###
#
# 1. Rewrites replay request by default
# 2. BEFORE_REPLAY gets triggered
#
def handle_request_replay_without_rewrite(replay_context: ReplayContext):
    options = { 'no_rewrite': True }
    handle_request_replay(replay_context, **options)

def handle_request_replay(replay_context: ReplayContext, **options: ReplayOptions):
    request: MitmproxyRequest = replay_context.flow.request
    intercept_settings: InterceptSettings = replay_context.intercept_settings

    policy = get_active_mode_policy(request, intercept_settings)
    if policy != replay_policy.NONE:
        if not options.get('no_rewrite'):
            __rewrite_request(replay_context)

        __replay_hook(lifecycle_hooks.BEFORE_REPLAY, replay_context)

###
#
# 1. Rewrites replay response
# 2. AFTER_REPLAY gets triggered
#
def handle_response_replay(replay_context: ReplayContext):
    __rewrite_response(replay_context)
    __replay_hook(lifecycle_hooks.AFTER_REPLAY, replay_context)

def __replay_hook(hook: str, replay_context: ReplayContext):
    intercept_settings: InterceptSettings = replay_context.intercept_settings

    lifecycle_hooks_module = intercept_settings.lifecycle_hooks
    if hook in lifecycle_hooks_module:
        lifecycle_hooks_module[hook](replay_context)

def __rewrite_request(replay_context: ReplayContext):
    """
    Before replaying a request, see if the request needs to be rewritten
    """
    intercept_settings: InterceptSettings = replay_context.intercept_settings
    rewrite_rules = intercept_settings.replay_rewrite_rules

    if len(rewrite_rules) > 0:
        rewrite_request(replay_context.flow, rewrite_rules)

def __rewrite_response(replay_context: ReplayContext):
    """
    After replaying a request, see if the request needs to be rewritten
    """
    intercept_settings: InterceptSettings = replay_context.intercept_settings
    rewrite_rules = intercept_settings.replay_rewrite_rules

    if len(rewrite_rules) > 0:
        rewrite_response(replay_context.flow, rewrite_rules)


import json
import pdb

from stoobly_agent.config.data_dir import DataDir

class MitmproxyConfig():
  __instance = None
  __master = None

  def __init__(self, master):
    if self.__instance:
        raise RuntimeError('Call instance() instead')
    else:
        self.__master = master 

  @classmethod
  def instance(cls, master = None):
      if cls.__instance is None:
        cls.__instance = cls(master)

      return cls.__instance

  def with_master(self, master):
    from mitmproxy.tools.dump import DumpMaster
    self.__master: DumpMaster = master

  def get(self, key: str):
    if not self.__master:
      try:
        fp = open(DataDir.instance().mitmproxy_options_json_path, 'r')
        contents = fp.read()
        fp.close()
        options = json.loads(contents)
        return options.get(key)
      except Exception as e:
        pass
    else:
      from mitmproxy.options import Options
      options: Options = self.__master.options

      if key in options:
        for k, val in options.items():
          if key == k:
            return val.current()

  def set(self, option: tuple):
    if self.__master:
      self.__master.options.set(*option)

  def dump(self):
    if not self.__master:
      return

    options = {}
    for k, v in self.__master.options.items():
      options[k] = v.current()

    fp = open(DataDir.instance().mitmproxy_options_json_path, 'w')
    fp.write(json.dumps(options, indent=2, sort_keys=True))
    fp.close()


from typing import Dict, List, Literal, TypedDict

from ..constants.firewall_action import FirewallAction
from ..constants.intercept_mode import Mode
from ..constants.request_component import RequestComponent

Method = Literal['DELETE,GET,OPTIONS,POST,PUT']
MockPolicy = Literal['all', 'found', 'none']
RecordPolicy = Literal['all','found','none']
TestPolicy = Literal['all','found','none']

class ParameterRule(TypedDict):
  modes: List[Mode]
  name: str
  type: str
  value: str

class UrlRule(TypedDict):
  hostname: str
  modes: List[Mode]
  port: str
  scheme: str

class DataRules(TypedDict):
  mock_policy: MockPolicy
  record_policy: RecordPolicy
  scenario_id: str
  test_policy: TestPolicy

class MatchRule(TypedDict):
  components: List[RequestComponent]
  modes: List[Mode]
  pattern: str

class RewriteRule(TypedDict):
  methods: List[Method]
  pattern: str
  parameter_rules: List[ParameterRule]
  url_rule: List[UrlRule]

class FirewallRule(TypedDict):
  action: List[FirewallAction]
  modes: List[Mode]
  pattern: str

class InterceptSettings(TypedDict):
  active: bool
  mode: Mode
  project_id: str
  scenario_id: str
  upstream_url: str

DataSettings = Dict[str, DataRules]
MatchRules = Dict[str, List[MatchRule]]
MatchSettings = Dict[str, MatchRules]
RewriteRules = List[RewriteRule]
RewriteSettings = Dict[str, RewriteRules]
FirewallRules = Dict[str, List[FirewallRule]]
FirewallSettings = Dict[str, FirewallRules]

class ProxySettings(TypedDict):
  data: DataSettings
  filter: RewriteSettings
  firewall: FirewallSettings
  intercept: InterceptSettings
  match: MatchSettings
  url: str


import requests

from http.cookies import SimpleCookie
from urllib3.exceptions import InsecureRequestWarning

from stoobly_agent.app.api.simple_http_request_handler import SimpleHTTPRequestHandler
from stoobly_agent.config.constants import custom_headers, headers
from stoobly_agent.config.mitmproxy import MitmproxyConfig
from stoobly_agent.lib.logger import Logger

LOG_ID = 'ProxyController'

class ProxyController:
    _instance = None

    def __init__(self):
        if self._instance:
            raise RuntimeError('Call instance() instead')
        else:
            self.data = {}

    @classmethod
    def instance(cls):
        if cls._instance is None:
            cls._instance = cls()

        return cls._instance

    def do_DELETE(self, context):
        self.__proxy(context, requests.delete)

    def do_GET(self, context):
        self.__proxy(context, requests.get)

    def do_HEAD(self, context):
        self.__proxy(context, requests.head)

    def do_OPTIONS(self, context):
        self.__proxy(context, requests.options)

    def do_PATCH(self, context):
        self.__proxy(context, requests.patch)

    def do_POST(self, context):
        self.__proxy(context, requests.post)

    def do_PUT(self, context):
        self.__proxy(context, requests.put)

    def __proxy(self, context: SimpleHTTPRequestHandler, method):
        url = self.__get_url(context)

        if url:
            _headers = self.__get_headers(context)
            _cookies =  self.__get_cookies(context)
            _body = self.__get_body(context)
            _params = context.params
            _verify = not MitmproxyConfig.instance().get('ssl_insecure')

            Logger.instance(LOG_ID).debug('Request Headers')
            Logger.instance(LOG_ID).debug(_headers)
            Logger.instance(LOG_ID).debug('Cookies')
            Logger.instance(LOG_ID).debug(_cookies)
            Logger.instance(LOG_ID).debug('Body')
            Logger.instance(LOG_ID).debug(_body)
            Logger.instance(LOG_ID).debug('Query Params')
            Logger.instance(LOG_ID).debug(_params)

            body = None
            headers = {}
            status = 0

            if not _verify:
                # Suppress only the single warning from urllib3 needed.
                requests.packages.urllib3.disable_warnings(category=InsecureRequestWarning)

            try:
                res = method(
                    url,
                    allow_redirects = True,
                    cookies = _cookies,
                    data = _body,
                    headers = _headers,
                    params = _params,
                    stream = True,
                    verify = _verify
                )
                
                body = res.raw.data
                headers = res.headers
                status = res.status_code
            except requests.exceptions.ConnectTimeout:
                body = b'Gateway Timeout'
                status = 504
            except requests.exceptions.SSLError as e:
                body = str(e).encode()
                status = 502
            except requests.exceptions.ConnectionError:
                body = b'Bad Gateway'
                status = 502
            except Exception:
                body = b'Unknown Error'
                status = 0

            Logger.instance(LOG_ID).debug('Response Headers')
            Logger.instance(LOG_ID).debug(res.headers)

            context.render(
                headers = headers,
                data = body,
                status = status,
            )

    def __get_headers(self, context: SimpleHTTPRequestHandler):
        request_headers = dict(context.headers)

        headers_white_list = []

        if headers.PROXY_HEADERS.title() in request_headers:
            headers_white_list = request_headers[headers.PROXY_HEADERS.title()].split(',')

        white_listed_headers = {}
        for name, value in request_headers.items():
            if name in headers_white_list:
                white_listed_headers[name] = value

        return white_listed_headers

    def __get_url(self, context: SimpleHTTPRequestHandler):
        service_url = context.headers.get(custom_headers.SERVICE_URL)

        if not service_url:
            context.render(
                plain = f"Invalid {custom_headers.SERVICE_URL} header {service_url}",
                status = 400
            )
            return None

        request_path = context.headers.get(headers.REQUEST_PATH)

        if not request_path:
            context.render(
                plain = f"Invalid {headers.REQUEST_PATH} header {service_url}",
                status = 400
            )
            return None

        return f"{service_url}{request_path}"

    def __get_cookies(self, context: SimpleHTTPRequestHandler):
        return SimpleCookie(context.headers.get('Cookie'))

    def __get_body(self, context):
        content_length = context.headers.get('Content-Length')
        if content_length:
            return context.rfile.read(content_length)





from ...constants import SERVICE_DNS

from .builder import WorkflowBuilder

class DnsDecorator():

  def __init__(self, workflow_builder: WorkflowBuilder):
    self.__workflow_builder = workflow_builder

  @property
  def workflow_builder(self):
    return self.__workflow_builder

  @property
  def service_builder(self):
    return self.workflow_builder.service_builder

  def decorate(self):
    config = self.service_builder.config
    proxy_name = self.workflow_builder.proxy
    services = self.workflow_builder.services

    # If we are reverse proxying to potentially an external host,
    # Docker's embedded DNS will use the host's /etc/host file as part of the resolution process
    # This can lead the hostname to resolve to localhost instead of the service's actual IP address
    if not config.detached:
      if not proxy_name in services:
        services[proxy_name] = {}

      services[proxy_name]['dns'] = SERVICE_DNS


import copy
import pdb

from typing import Union

from stoobly_agent.app.proxy.test.matchers.context import MatchContext
from stoobly_agent.app.settings.constants import request_component
from stoobly_agent.config.constants.lifecycle_hooks import BEFORE_REQUEST_COMPONENT_CREATE, BEFORE_REQUEST_COMPONENT_DELETE

class SynchronizeEvent():

  def __init__(self, component_type: request_component.RequestComponent, previous_state: Union[dict, list], key, value):
    self.__component_type = component_type
    self.__key = key
    self.__previous_state = previous_state
    self.__value = value

  @property
  def key(self):
    return self.__key

  @property
  def previous_state(self):
    return self.__previous_state

  @property
  def value(self):
    return self.__value

  @property
  def is_body_param(self):
    return self.__component_type == request_component.BODY_PARAM

  @property
  def is_query_param(self):
    return self.__component_type == request_component.QUERY_PARAM

class SynchronizeCreateEvent(SynchronizeEvent):

  def __init__(self, component_type: request_component.RequestComponent, previous_state: Union[dict, list], key, value):
    super().__init__(component_type, previous_state, key, value)

class SynchronizeDeleteEvent(SynchronizeEvent):

  def __init__(self, component_type: request_component.RequestComponent, previous_state: Union[dict, list], key):
    super().__init__(component_type, previous_state, key, None)

class EventLog():

  def __init__(self):
    self.__events = []

  def append(self, event: SynchronizeEvent):
    self.__events.append(event)

class RequestSynchronizeHandler():

  def __init__(self, component_type: request_component.RequestComponent, lifecycle_hooks = {}):
    self.__component_type = component_type
    self.__event_log = EventLog()
    self.__lifecycle_hooks = lifecycle_hooks

  @property
  def event_log(self):
    return self.__event_log

  def handle_length_match_error(self, context: MatchContext, value: list):
    return True

  def handle_param_name_missing(self, context: MatchContext, value: dict):
    request_component_name = context.request_component_name

    # Component deleted already
    if not request_component_name:
      return True

    potential_values = request_component_name.get('values') or [None]

    if len(potential_values) > 0:
      key = context.current_key
      new_value = potential_values[0]

      event = SynchronizeCreateEvent(self.__component_type, copy.deepcopy(value), key, new_value)
      handle_before_param_create = self.__lifecycle_hooks.get(BEFORE_REQUEST_COMPONENT_CREATE)

      if not handle_before_param_create or not handle_before_param_create(event):
        value[key] = new_value
        self.event_log.append(event)

    return True

  def handle_param_name_exists(self, context: MatchContext, value: dict):
    key = context.current_key

    event = SynchronizeDeleteEvent(self.__component_type, copy.deepcopy(value), key)
    handle_before_param_delete = self.__lifecycle_hooks.get(BEFORE_REQUEST_COMPONENT_DELETE)

    if not handle_before_param_delete or not handle_before_param_delete(event):
      del value[key]
      self.event_log.append(event)

    return True

  def handle_type_match_error(self, context: MatchContext, value):
    if isinstance(value, list) or isinstance(value, dict):
      key = context.current_key
      del value[key]

    return True

  def handle_valid_type_error(self, context: MatchContext, value):
    if isinstance(value, list) or isinstance(value, dict):
      key = context.current_key
      del value[key]
      
    return True



import pdb
import urllib

from ..logger import Logger
from .interfaces import RequestCreateParams, RequestsIndexQueryParams, RequestShowQueryParams
from .stoobly_api import StooblyApi

LOG_ID = 'RequestsResource'

class RequestsResource(StooblyApi):

  def create(self, **body_params: RequestCreateParams):
      url = f"{self.service_url}{self.REQUESTS_ENDPOINT}"

      requests = body_params.get('requests')

      if requests:
        del body_params['requests']

      proxies = {}

      return self.post(url, headers=self.default_headers, data=body_params, files={ 'requests': requests }, proxies=proxies)

  def index(self, **query_params: RequestsIndexQueryParams):
    url = f"{self.service_url}{self.REQUESTS_ENDPOINT}"

    Logger.instance(LOG_ID).debug(f"{url}?{urllib.parse.urlencode(query_params)}")

    return self.get(url, headers=self.default_headers, params=query_params)

  def response(self, **query_params):
    url = f"{self.service_url}{self.REQUESTS_ENDPOINT}/response"

    Logger.instance(LOG_ID).debug(f"{url}?{urllib.parse.urlencode(query_params)}")

    return self.get(
        url,
        allow_redirects=False,
        headers=self.default_headers,
        params=query_params,
        stream=True
    )

  def show(self, request_id, **query_params: RequestShowQueryParams):
      url = f"{self.service_url}{self.REQUESTS_ENDPOINT}/{request_id}"

      Logger.instance(LOG_ID).debug(f"{url}?{urllib.parse.urlencode(query_params)}")

      return self.get(url, headers=self.default_headers, params=query_params)




import pdb
import requests

from stoobly_agent.app.cli.helpers.context import ReplayContext
from stoobly_agent.app.settings import Settings
from stoobly_agent.config.constants import custom_headers
from stoobly_agent.lib.api.test_responses_resource import TestResponsesResource
from stoobly_agent.lib.api.tests_resource import TestsResource
from stoobly_agent.lib.api.interfaces import TestCreateParams, TestShowResponse
from stoobly_agent.lib.api.keys import OrganizationKey, TestKey

class TestFacade():

  def __init__(self, __settings: Settings):
    self.__settings = __settings
    self.__api = TestsResource(self.__settings.remote.api_url, self.__settings.remote.api_key)

  def create(self, **kwargs: TestCreateParams) -> TestShowResponse:
    organization_key: str = OrganizationKey(kwargs.get('organization_key')) 

    res: requests.Response =  self.__api.create(
      description=kwargs.get('description') or '',
      name=kwargs.get('name'),
      organization_id=organization_key.id
    )

    if not res.ok:
      raise AssertionError(res.content)

    return res.json()

  def expected_repsonse(self, test_key: str) -> requests.Response:
    api = TestResponsesResource(self.__settings.remote.api_url, self.__settings.remote.api_key)
    key = TestKey(test_key)
    res = api.mock(key.id, **{ 'expected': True, 'project_id': key.project_id })

    if not res.ok:
      raise AssertionError(res.content)

    return res

  def show(self, test_key: str) -> TestShowResponse:
    key = TestKey(test_key)
    res = self.__api.show(key.id)

    if not res.ok:
      raise AssertionError(res.content)

    return res.json()

  def show_with_context(self, context: ReplayContext, project_id: str):
    response = context.response

    test_id = response.headers.get(custom_headers.TEST_ID)
    if test_id:
        test_key = TestKey.encode({'i': test_id, 'p': project_id})
        return self.show(test_key)

  def expected_response_with_context(self, context: ReplayContext, project_id: str):
    response = context.response

    test_id = response.headers.get(custom_headers.TEST_ID)
    if test_id:
        test_key = TestKey.encode({'i': test_id, 'p': project_id})
        return self.expected_repsonse(test_key)


import pdb

from stoobly_agent.app.models.adapters.python.response import MitmproxyResponseAdapter as PythonResponseMitmproxyResponseAdapter
from stoobly_agent.app.models.adapters.raw_http_response_adapter import RawHttpResponseAdapter
from stoobly_agent.lib.orm.response import Response

class MitmproxyResponseAdapter():

  def __init__(self, response: Response):
    self.__response = response

  def adapt(self):
    python_response = RawHttpResponseAdapter(self.__response.raw).to_response()
    return PythonResponseMitmproxyResponseAdapter(python_response).adapt()


import pdb

from typing import Tuple

from stoobly_agent.app.models.types import ScenarioCreateParams
from stoobly_agent.lib.api.scenarios_resource import ScenariosResource
from stoobly_agent.lib.api.interfaces.scenarios import ScenarioShowResponse, ScenariosIndexResponse

class StooblyScenarioAdapter():

  def __init__(self, __api: ScenariosResource):
    self.__api = __api

  def create(self, **params: ScenarioCreateParams) -> Tuple[ScenarioShowResponse, int]:
    res = self.__api.create(**params)
    res.raise_for_status()  
    return res.json(), res.status_code

  def show(self, scenario_id: str) -> Tuple[ScenarioShowResponse, int]:
    res = self.__api.show(scenario_id)
    res.raise_for_status()  
    return res.json(), res.status_code

  def index(self, **query_params) -> Tuple[ScenariosIndexResponse, int]:
    res = self.__api.index(**query_params)
    res.raise_for_status()  
    return res.json(), res.status_code

  def update(self, scenario_id: str, **params):
    res = self.__api.update(scenario_id, **params)
    res.raise_for_status()  
    return res.json(), res.status_code


from typing import List, TypedDict

from stoobly_agent.lib.api.interfaces.headers import Header
from stoobly_agent.lib.api.interfaces.query_params import QueryParam
from stoobly_agent.lib.api.interfaces.response_headers import ResponseHeader

class HeaderIndexResponse(TypedDict):
  list: List[Header]

class QueryParamIndexResponse(TypedDict):
  list: List[QueryParam]
  total: int

class ResponseHeaderIndexResponse(TypedDict):
  list: List[ResponseHeader]
  total: int


import json
import pdb
import pytest

from click.testing import CliRunner
from typing import List

from stoobly_agent.test.test_helper import DETERMINISTIC_GET_REQUEST_URL, reset

from stoobly_agent.app.cli.helpers.handle_replay_service import JSON_FORMAT
from stoobly_agent.app.models.adapters.raw_http_response_adapter import RawHttpResponseAdapter
from stoobly_agent.app.cli.types.output import ReplayOutput
from stoobly_agent.cli import record, scenario
from stoobly_agent.lib.orm.request import Request
from stoobly_agent.lib.orm.scenario import Scenario
from stoobly_agent.lib.utils.decode import decode

@pytest.fixture(scope='module')
def runner():
    return CliRunner()

class TestReplay():
    @pytest.fixture(scope='class', autouse=True)
    def settings(self):
        return reset()

    class TestWhenNoOptions():

        @pytest.fixture(scope='class')
        def created_scenario(self, runner: CliRunner):
            create_result = runner.invoke(scenario, ['create', 'test'])
            assert create_result.exit_code == 0
            return Scenario.last()

        @pytest.fixture(scope='class')
        def recorded_request(self, runner: CliRunner, created_scenario: Scenario):
            record_result = runner.invoke(record, ['--scenario-key', created_scenario.key(), DETERMINISTIC_GET_REQUEST_URL])
            assert record_result.exit_code == 0
            return Request.last()

        @pytest.fixture(scope='class')
        def recorded_response_raw(self, recorded_request: Request):
            return recorded_request.response.raw
            
        def test_it_matches_recorded_response(self, runner, created_scenario: Scenario, recorded_response_raw: bytes):
            replay_result = runner.invoke(scenario, ['replay', '--format', JSON_FORMAT, created_scenario.key()])
            assert replay_result.exit_code == 0

            try:
                replay_outputs: List[ReplayOutput] = json.loads(replay_result.stdout)
            except Exception:
                assert False, 'Invalid JSON output'
            
            replay_output = replay_outputs[0]
            assert not not replay_output.get('content')

            response = RawHttpResponseAdapter(recorded_response_raw).to_response()
            assert replay_output['content']  == decode(response.content)


        def test_it_does_not_override_response_raw(self, recorded_request: Request, recorded_response_raw: bytes):
            new_response = Request.find(recorded_request.id).response.raw
            assert new_response == recorded_response_raw

    class TestWhenOverwritting():

        @pytest.fixture(scope='class')
        def created_scenario(self, runner: CliRunner):
            create_result = runner.invoke(scenario, ['create', 'test'])
            assert create_result.exit_code == 0
            return Scenario.last()

        @pytest.fixture(scope='class')
        def recorded_request(self, runner: CliRunner, created_scenario: Scenario):
            record_result = runner.invoke(record, ['--scenario-key', created_scenario.key(), DETERMINISTIC_GET_REQUEST_URL])
            assert record_result.exit_code == 0
            return Request.last()

        @pytest.fixture(scope='class')
        def recorded_response_raw(self, recorded_request: Request):
            return recorded_request.response.raw

        def test_it_overwrites_response_raw(self, runner: CliRunner, created_scenario: Scenario, recorded_request: Request, recorded_response_raw: bytes):
            replay_result = runner.invoke(scenario, ['replay', '--overwrite', created_scenario.key()])
            assert replay_result.exit_code == 0

            new_response = Request.find(recorded_request.id).response.raw
            assert new_response != recorded_response_raw

        def test_it_creates_one_replayed_response(self, recorded_request: Request):
            assert len(recorded_request.replayed_responses) == 1

        def test_it_sets_replayed_response_raw_to_response_raw(self, recorded_request: Request, recorded_response_raw: bytes):
            replayed_response = recorded_request.replayed_responses[0]
            assert replayed_response.raw == recorded_response_raw

    class TestWhenRecord():

        @pytest.fixture(scope='class')
        def created_scenario(self, runner: CliRunner):
            create_result = runner.invoke(scenario, ['create', 'test'])
            assert create_result.exit_code == 0
            return Scenario.last()

        @pytest.fixture(scope='class', autouse=True)
        def recorded_request(self, runner: CliRunner, created_scenario: Scenario):
            record_result = runner.invoke(record, ['--scenario-key', created_scenario.key(), DETERMINISTIC_GET_REQUEST_URL])
            assert record_result.exit_code == 0
            return Request.last()

        def test_it_records(self, runner: CliRunner, created_scenario: Scenario):
            replay_result = runner.invoke(scenario, ['replay', '--record', created_scenario.key()])
            assert replay_result.exit_code == 0

            assert Request.count() == 2

            _scenario = Scenario.find(created_scenario.id)
            assert _scenario.requests.count() == 1

        class TestWhenScenario():
            @pytest.fixture(scope='class', autouse=True)
            def settings(self):
                return reset()

            @pytest.fixture(scope='class')
            def source_scenario(self, runner: CliRunner):
                create_result = runner.invoke(scenario, ['create', 'test'])
                assert create_result.exit_code == 0
                return Scenario.last()

            @pytest.fixture(scope='class')
            def destination_scenario(self, runner: CliRunner):
                create_result = runner.invoke(scenario, ['create', 'test'])
                assert create_result.exit_code == 0
                return Scenario.last()

            @pytest.fixture(scope='class', autouse=True)
            def recorded_request(self, runner: CliRunner, source_scenario: Scenario):
                record_result = runner.invoke(record, ['--scenario-key', source_scenario.key(), DETERMINISTIC_GET_REQUEST_URL])
                assert record_result.exit_code == 0
                return Request.last()

            def test_it_records_to_scenario(self, runner: CliRunner, source_scenario: Scenario, destination_scenario: Scenario):
                replay_result = runner.invoke(scenario, ['replay', '--record' , '--scenario-key', destination_scenario.key(), source_scenario.key()])
                assert replay_result.exit_code == 0

                assert Request.count() == 2

                _scenario = Scenario.find(source_scenario.id)
                assert _scenario.requests.count() == 1

                _scenario = Scenario.find(destination_scenario.id)
                assert _scenario.requests.count() == 1


import json
import mimetypes
import os
import pdb
import requests

from http.server import BaseHTTPRequestHandler

class SimpleHTTPRequestHandler(BaseHTTPRequestHandler):
    def render(self, **kwargs):
        if kwargs.get('file') != None:
            self.render_file(kwargs)
        elif kwargs.get('json') != None:
            self.render_json(kwargs)
        elif kwargs.get('plain') != None:
            self.render_plain(kwargs)
        elif kwargs.get('data') != None:
            self.render_data(kwargs)
        elif kwargs.get('download') != None:
            self.render_download(kwargs)

    def render_file(self, kwargs):
        if not 'headers' in kwargs:
            kwargs['headers'] = {}

        path = kwargs['file']

        if not os.path.exists(path):
            kwargs['status'] = 404

        mimetype = mimetypes.guess_type(path)[0]
        kwargs['headers']['Content-Type'] = mimetype or 'text/plain'

        fp = open(path, 'rb')
        kwargs['data'] = fp.read()
        self.render_data(kwargs)

    def render_json(self, kwargs):
        if not 'headers' in kwargs:
            kwargs['headers'] = {}

        kwargs['headers']['Content-Type'] = 'application/json'
        kwargs['data'] = json.dumps(kwargs['json'])
        self.render_data(kwargs)

    def render_plain(self, kwargs):
        if not 'headers' in kwargs:
            kwargs['headers'] = {}

        kwargs['headers']['Content-Type'] = 'text/plain'
        kwargs['data'] = kwargs['plain']
        self.render_data(kwargs)


    def render_download(self, kwargs):
        if not 'headers' in kwargs:
            kwargs['headers'] = {}

        kwargs['headers']['Content-Disposition'] = f"attachment; filename=\"{kwargs.get('filename') or 'download.txt'}\""
        kwargs['data'] = kwargs['download']
        self.render_data(kwargs)

    def render_data(self, kwargs):
        self.send_response(kwargs.get('status') or 200)

        body = b''
        if isinstance(kwargs['data'], str):
            body = kwargs['data'].encode()
        else:
            body = kwargs['data']

        # Send headers
        headers = self.filter_headers(kwargs.get('headers'), {
            'TRANSFER-ENCODING': 'CHUNKED',
        })
        headers['Content-Length'] = str(len(body))

        rendered_headers = self.enable_cors(headers)
        self.render_headers(headers, rendered_headers)
        self.end_headers()

        # Send body
        self.wfile.write(body)

    def filter_response(self, res, status):
        if status < 400:
            return False

        self.render(
            plain = res,
            status = status 
        )

        return True

    def pass_on(self, res: requests.Response):
        self.render(
            headers = res.headers,
            data = res.raw.data,
            status = res.status_code,
        )

    def bad_request(self, message = ''):
        self.render(
            plain = message,
            status = 400
        )

    def not_found(self, message = ''):
        self.render(
            plain = message,
            status = 404
        )

    def internal_error(self, message = ''):
        self.render(
            plain = message,
            status = 500
        )

    def filter_headers(self, headers, blacklist = {}):
        if not headers:
            return {}

        new_blacklist = {}
        for key, val in blacklist.items():
            new_blacklist[key.upper()] = val

        new_headers = headers.copy()
        for key, val in headers.items():
            if key.upper() in new_blacklist:
                if not new_blacklist[key.upper()]:
                    del new_headers[key.upper()]
                elif val.upper() == new_blacklist[key.upper()]:
                    del new_headers[key.upper()]

        return new_headers


    def render_headers(self, headers, ignored_headers = []):
        if headers:
            for key, val in headers.items():
                if key in ignored_headers:
                    continue

                self.send_header(key, val)

    def required_params(self, params: dict, required_params: list):
        if not isinstance(params, dict):
            self.not_found(f"Body is empty")
            return False

        for param in required_params:
            if param not in params:
                self.bad_request(f"Missing param {param}")
                return False

        return True

    def enable_cors(self):
        pass


import pdb

from stoobly_agent.app.api.simple_http_request_handler import SimpleHTTPRequestHandler
from stoobly_agent.app.models.response_header_model import ResponseHeaderModel
from stoobly_agent.app.settings import Settings

class ResponseHeadersController:
    _instance = None

    def __init__(self):
        if self._instance:
            raise RuntimeError('Call instance() instead')
        else:
            self.data = {}

    @classmethod
    def instance(cls):
        if cls._instance is None:
            cls._instance = cls()

        return cls._instance

    # POST /requests/:requestId/response_headers
    def create(self, context: SimpleHTTPRequestHandler):
        context.parse_path_params({
            'requestId': 1
        })

        response_header_model = self.__response_header_model(context)
        header, status = response_header_model.create(context.params.get('requestId'), **context.params)

        if context.filter_response(header, status):
            return
        
        context.render(
            json = header,
            status = 200
        )

    # GET /requests/:requestId/headers
    def index(self, context: SimpleHTTPRequestHandler):
        context.parse_path_params({
            'requestId': 1
        })

        response_header_model = self.__response_header_model(context)
        response_headers, status = response_header_model.index(context.params.get('requestId'), **context.params)

        if context.filter_response(response_headers, status):
            return

        context.render(
            json = response_headers,
            status = 200
        )

    # PUT /requests/:requestId/response_headers/:headerId
    def update(self, context: SimpleHTTPRequestHandler):
        context.parse_path_params({
            'requestId': 1
        })

        response_header_model = self.__response_header_model(context)
        header, status = response_header_model.update(context.params.get('requestId'), **context.params)

        if context.filter_response(header, status):
            return
        
        context.render(
            json = header,
            status = 200
        )

    # PUT /requests/:requestId/response_headers/:responseHeaderId
    def destroy(self, context: SimpleHTTPRequestHandler):
        context.parse_path_params({
            'requestId': 1,
            'responseHeaderId': 3,
        })

        response_header_model = self.__response_header_model(context)
        header, status = response_header_model.destroy(context.params.get('requestId'), context.params.get('responseHeaderId'))

        if context.filter_response(header, status):
            return

        context.render(
            plain = '',
            status = 200
        )

    def __response_header_model(self, context: SimpleHTTPRequestHandler):
        access_token = context.headers.get('access-token')
        response_header_model = ResponseHeaderModel(Settings.instance(), access_token=access_token)
        return response_header_model


from requests import Response

from ..context_response import TestContextResponse

class RequestsResponseAdapter():

  def __init__(self, response: Response):
    self._response = response

  def adapt(self):
    response = TestContextResponse()

    response.with_content(self._response.content)
    response.with_headers(self._response.headers)
    response.with_status_code(self._response.status_code)

    return response


import pdb
import pytest

from typing import List

from stoobly_agent.app.proxy.replay.alias_resolver import AliasResolver
from stoobly_agent.app.proxy.replay.rewrite_params_service import rewrite_params
from stoobly_agent.config.constants import alias_resolve_strategy
from stoobly_agent.lib.orm.trace import Trace
from stoobly_agent.lib.orm.trace_alias import TraceAlias

@pytest.fixture
def trace():
  return Trace.create()

class Test():

  def test_rewrites_dict(self, trace):
    response = {
      'id': 1,
    }

    response_param_names = [
      {
        'alias_id': 1,
        'name': 'id',
        'query': 'id',
      }
    ]

    id_to_alias = {
      1: {
        'name': ':requestId',
      }
    } 

    # Create trace alias
    resolved_alias_values = [1000]
    self.build_trace_aliases(trace, id_to_alias.values(), resolved_alias_values)

    alias_resolver = AliasResolver(trace, alias_resolve_strategy.LIFO)
    rewrite_params(response, response_param_names, id_to_alias, alias_resolver)

    expected_id = resolved_alias_values[0]
    assert response['id'] == expected_id, self.value_not_match_error(response['id'], expected_id)

  def test_rewrites_list_of_dicts(self, trace):
    response = [{
      'id': 1,
    }]

    response_param_names = [
      {
        'alias_id': 1,
        'name': 'id',
        'query': '[*].id',
      }
    ]

    id_to_alias = {
      1: {
        'name': ':requestId',
      }
    } 

    # Create trace alias
    resolved_alias_values = [1000]
    self.build_trace_aliases(trace, id_to_alias.values(), resolved_alias_values)

    alias_resolver = AliasResolver(trace, alias_resolve_strategy.LIFO)
    rewrite_params(response, response_param_names, id_to_alias, alias_resolver)

    expected_id = resolved_alias_values[0]
    assert response[0]['id'] == expected_id, self.value_not_match_error(response['id'], expected_id)

  def test_rewrites_dict_of_list_of_dicts_of_dicts(self, trace):
    response = {'list': [{'id': 910, 'service_id': 26, 'name': 'users', 'position': 0, 'created_at': '2022-10-03T00:59:15.172Z', 'updated_at': '2022-10-03T00:59:15.172Z'}, {'id': 911, 'service_id': 26, 'name': ':userId2', 'position': 1, 'created_at': '2022-10-03T00:59:15.404Z', 'updated_at': '2022-10-03T00:59:20.238Z', 'alias_name': ':userId2', 'alias': {'id': 144, 'name': ':userId2', 'created_at': '2022-07-28T08:16:59.599Z', 'updated_at': '2022-08-22T08:47:54.567Z', 'project_id': 38, 'path_segment_names_count': -1, 'header_names_count': 0, 'query_param_names_count': 0, 'body_param_names_count': 0, 'response_param_names_count': 0, 'service_id': 26}}], 'total': 2}
    response_param_names = [{'id': 23993, 'alias_id': 152, 'endpoint_id': 1477, 'response_param_name_id': 23991, 'name': 'id', 'query': 'list[*].alias.id', 'inferred_type': 'Integer', 'response_param_names_count': 0, 'is_required': True, 'created_at': '2022-08-25T06:53:31.663Z', 'updated_at': '2022-08-25T06:54:51.936Z', 'requests_count': 4, 'is_deterministic': False}]
    id_to_alias = {152: {'id': 152, 'name': ':aliasId'}}

    resolved_alias_values = [144]
    self.build_trace_aliases(trace, id_to_alias.values(), resolved_alias_values)

    alias_resolver = AliasResolver(trace, alias_resolve_strategy.LIFO)
    rewrite_params(response, response_param_names, id_to_alias, alias_resolver)

    expected_id = resolved_alias_values[0]
    assert response['list'][1]['alias']['id'] == expected_id, self.value_not_match_error(response['id'], expected_id)

  def build_trace_aliases(self, trace: Trace, aliases: List[TraceAlias], values: list):
    for i, _alias in enumerate(aliases):
      TraceAlias.create(name=_alias['name'], trace_id=trace.id, value=values[i])

  def value_not_match_error(self, v1, v2):
    return f"Got {v1}, expected {v2}"


import pdb
import uuid

from stoobly_orator.orm import has_many

from stoobly_agent.lib import orm
from stoobly_agent.lib.api.keys.project_key import LOCAL_PROJECT_ID
from stoobly_agent.lib.api.keys.scenario_key import ScenarioKey

from .base import Base

class Scenario(Base):
  __fillable__ = ['name','description', 'is_deleted', 'overwritable', 'position', 'priority', 'starred', 'uuid']
  
  @has_many
  def requests(self):
    return Request

  def key(self):
    return ScenarioKey.encode(LOCAL_PROJECT_ID, self.uuid)

  # Override
  def to_dict(self):
    h = super().attributes_to_dict()
    h['key'] = self.key()
    return h

def handle_creating(scenario):
  if not hasattr(scenario, 'uuid') or not scenario.uuid:
    scenario.uuid = str(uuid.uuid4())
  else:
    try:
      uuid.UUID(scenario.uuid)
    except Exception as e:
      scenario.uuid = str(uuid.uuid4())

def handle_deleting(scenario):
  requests = scenario.requests
  
  for request in requests:
    request.delete()

Scenario.creating(handle_creating)
Scenario.deleting(handle_deleting)

from .request import Request


from typing import List, Union 

from stoobly_agent.config.constants import mode
from stoobly_agent.config.constants import mock_policy
from stoobly_agent.config.constants import record_policy
from stoobly_agent.config.constants import test_strategy

class ActiveModeSettingsBuilder():
  exclude_patterns: list = []
  include_patterns: list = []
  policy: str = None
  project_key: str = None
  rewrite_rules: list = []
  service_url: str = None
  scenario_key: str = None
  test_strategy: str = None
  enabled: bool  = False

  def __init__(self, active_mode: Union[mode.MOCK, mode.RECORD, mode.TEST]):
    if active_mode == mode.MOCK:
      self.policy = mock_policy.ALL
    elif active_mode == mode.RECORD:
      self.policy = record_policy.ALL
    elif active_mode == mode.TEST:
      self.policy = mock_policy.ALL
      self.test_strategy = test_strategy.DIFF

  def with_exclude_patterns(self, patterns: List[str]):
    self.exclude_patterns = patterns
    return self

  def with_include_patterns(self, patterns: List[str]):
    self.include_patterns = patterns
    return self

  def with_policy(self, policy: str):
    self.policy = policy
    return self

  def with_project_key(self, project_key: str):
    self.project_key = project_key
    return self

  def with_rewrite_rules(self, rules: List[str]):
    self.rewrite_rules = rules
    return self

  def with_service_url(self, service_url: str):
    self.service_url = service_url
    return self
  
  def with_scenario_key(self, scenario_key: str):
    self.scenario_key = scenario_key
    return self

  def with_test_strategy(self, test_strategy: str):
    self.test_strategy = test_strategy
    return self

  def with_enabled(self, enabled: bool):
    self.enabled = enabled
    return self

  def build(self):
    return {
      'exclude_patterns' : self.exclude_patterns,
      'include_patterns' : self.include_patterns,
      'policy' : self.policy,
      'project_key': self.project_key,
      'rewrite_rules': self.rewrite_rules,
      'service_url': self.service_url,
      'scenario_key': self.scenario_key,
      'test_strategy': self.test_strategy,
      'enabled': self.enabled,
    }


import pdb
import requests

from stoobly_agent.app.models.adapters import JoinedRequestAdapter, RawHttpRequestAdapter, RawHttpResponseAdapter
from stoobly_agent.app.models.adapters.python import PythonRequestAdapterFactory, PythonResponseAdapterFactory

from stoobly_agent.app.proxy.record.join_request_service import InterceptSettings, join_request, MitmproxyRequestFacade, MitmproxyResponseFacade
from stoobly_agent.app.settings import Settings
from stoobly_agent.lib.logger import Logger

LOG_ID = 'CreateRequestParamsService'

class MitmproxyFlowMock():
    def __init__(self, request, response):
        self.request = request
        self.response = response

def build_params(raw_requests: str, payloads_delimitter = None):
  try:
    joined_request = JoinedRequestAdapter(raw_requests, payloads_delimitter).adapt()
  except Exception as e:
    Logger.instance(LOG_ID).error(e)
    return

  request_adapter = RawHttpRequestAdapter(joined_request.request_string.get())
  response_adapter = RawHttpResponseAdapter(joined_request.response_string.get())

  mitmproxy_request = PythonRequestAdapterFactory(request_adapter.to_request()).mitmproxy_request(request_adapter.protocol)
  mitmproxy_response = PythonResponseAdapterFactory(response_adapter.to_response()).mitmproxy_response()

  mitmproxy_flow_mock = MitmproxyFlowMock(mitmproxy_request, mitmproxy_response)

  return {
    'flow': mitmproxy_flow_mock,
    'joined_request': joined_request,
  }

def build_params_from_python(request: requests.Request, response: requests.Response):
  mitmproxy_request = PythonRequestAdapterFactory(request).mitmproxy_request()
  mitmproxy_response = PythonResponseAdapterFactory(response).mitmproxy_response()

  intercept_settings = InterceptSettings(Settings.instance(), mitmproxy_request)
  joined_request = join_request(MitmproxyRequestFacade(mitmproxy_request), MitmproxyResponseFacade(mitmproxy_response), intercept_settings)

  mitmproxy_flow_mock = MitmproxyFlowMock(mitmproxy_request, mitmproxy_response)

  return {
    'flow': mitmproxy_flow_mock,
    'joined_request': joined_request,
  }


import os

from .app import App
from .workflow import Workflow

class ServiceWorkflow(Workflow):

  def __init__(self, service_name: str, workflow_name: str, app: App):
    super().__init__(workflow_name, app)
    self._service_name = service_name

  @property
  def path(self):
    return os.path.join(self.app.scaffold_namespace_path, self.service_name, self.workflow_name)

  @property
  def service_name(self):
    return self._service_name


import click
import json
import os
import pdb
import requests
import sys

from stoobly_agent import VERSION
from stoobly_agent.app.cli.helpers.context import ReplayContext
from stoobly_agent.app.cli.helpers.handle_mock_service import print_raw_response, RAW_FORMAT
from stoobly_agent.app.cli.helpers.validations import validate_project_key, validate_scenario_key
from stoobly_agent.app.proxy.constants import custom_response_codes
from stoobly_agent.app.proxy.replay.replay_request_service import replay as replay_request
from stoobly_agent.config.constants import env_vars, mode
from stoobly_agent.config.data_dir import DataDir
from stoobly_agent.lib.utils.conditional_decorator import ConditionalDecorator

from .app.api import initialize as initialize_api, run as run_api
from .app.cli import ca_cert, config, endpoint, feature, intercept, MainGroup, request, scenario, scaffold, snapshot, trace
from .app.cli.helpers.feature_flags import local, remote
from .app.settings import Settings
from .lib import logger
from .lib.orm.migrate_service import migrate as migrate_database
from .lib.utils.decode import decode

settings: Settings = Settings.instance()
is_remote = remote(settings)
is_local = local(settings)

# Makes sure database is up to date
migrate_database(VERSION)

CONNECTION_STRATEGIES = ['eager', 'lazy']
CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])

@click.version_option()
@click.group(
    cls=MainGroup,
    context_settings=CONTEXT_SETTINGS,
    epilog="Run 'stoobly-agent COMMAND --help' for more information on a command.",
)
@click.pass_context
def main(ctx):
  ctx.terminal_width = 256

# Attach subcommands to main
main.add_command(ca_cert)
main.add_command(config)
main.add_command(endpoint)
main.add_command(feature)
main.add_command(intercept)
main.add_command(request)
main.add_command(scaffold)
main.add_command(scenario)
main.add_command(snapshot)
main.add_command(trace)

if settings.cli.features.dev_tools:
    from .app.cli import dev_tools
    main.add_command(dev_tools)

if settings.cli.features.exec:
    from .app.cli.decorators.exec import ExecDecorator
    ExecDecorator(main).decorate()

if settings.cli.features.remote:
    from .app.cli import project, report
    main.add_command(project)
    main.add_command(report)

@main.command(
    help="Initialize a new context"
)
def init(**kwargs):
    DataDir.instance().create()
    settings.reset_and_load()

@main.command(
    help="Run proxy and/or UI",
)
@ConditionalDecorator(lambda f: click.option('--api-url', help='API URL.')(f), is_remote)
@click.option('--ca-certs-dir-path', default=DataDir.instance().ca_certs_dir_path, help='Path to ca certs directory used to sign SSL certs.')
@click.option('--certs', help='''
  SSL certificates of the form "[domain=]path". The domain may include a wildcard, and is equal to "*" if not specified. The file at path is a certificate in PEM format. If a private key is included in the
  PEM, it is used, else the default key in the conf dir is used. The PEM file should contain the full certificate chain, with the leaf certificate as the first entry. May be passed multiple times.
''')
@click.option('--cert-passphrase', help='''
  Passphrase for decrypting the private key provided in the --cert option. Note that passing cert_passphrase on the command line makes your passphrase visible in your system's process list. Specify it in
  config.yaml to avoid this.
''')
@click.option('--connection-strategy', help=', '.join(CONNECTION_STRATEGIES), type=click.Choice(CONNECTION_STRATEGIES))
@click.option('--flow-detail', default='1', type=click.Choice(['0', '1', '2', '3', '4']), help='''
  The display detail level for flows in mitmdump: 0 (quiet) to 4 (very verbose).
  0: no output
  1: shortened request URL with response status code
  2: full request URL with response status code and HTTP headers
  3: 2 + truncated response content, content of WebSocket and TCP messages (content_view_lines_cutoff: 512)
  4: 3 + nothing is truncated
''')
@click.option('--headless', is_flag=True, default=False, help='Disable starting UI.')
@click.option('--intercept', is_flag=True, default=False, help='Enable intercept on run.')
@click.option('--log-level', default=logger.INFO, type=click.Choice([logger.DEBUG, logger.INFO, logger.WARNING, logger.ERROR]), help='''
    Log levels can be "debug", "info", "warning", or "error"
''')
@click.option('--lifecycle-hooks-path', help='Path to lifecycle hooks script.')
@click.option('--modify-headers', multiple=True, help='''
  Header modify pattern of the form "[/flow-filter]/header-name/[@]header-value", where the separator can be any character. The @ allows to provide a file path that is used to read the header value string.
  An empty header-value removes existing header-name headers. May be passed multiple times.
''')
@click.option('--proxy-host', default='0.0.0.0', help='Address to bind proxy to.')
@click.option('--proxyless', is_flag=True, default=False, help='Disable starting proxy.')
@click.option('--proxy-mode', default="regular", help='''
  Proxy mode can be "regular", "transparent", "socks5",
  "reverse:SPEC", or "upstream:SPEC". For reverse and
  upstream proxy modes, SPEC is host specification in
  the form of "http[s]://host[:port]".
''')
@click.option('--proxy-port', default=8080, help='Proxy service port.')
@click.option('--public-directory-path', help='Path to public files. Used for mocking requests.')
@click.option('--response-fixtures-path', help='Path to response fixtures yaml. Used for mocking requests.')
@click.option('--ssl-insecure', is_flag=True, default=False, help='Do not verify upstream server SSL/TLS certificates.')
@click.option('--ui-host', default='0.0.0.0', help='Address to bind UI to.')
@click.option('--ui-port', default=4200, help='UI service port.')
def run(**kwargs):
    from .app.proxy.run import run as run_proxy

    os.environ[env_vars.AGENT_PROXY_URL] = f"http://{kwargs['proxy_host']}:{kwargs['proxy_port']}"

    if kwargs.get('lifecycle_hooks_path'):
      os.environ[env_vars.AGENT_LIFECYCLE_HOOKS_PATH] = kwargs['lifecycle_hooks_path']

    if kwargs.get('public_directory_path'):
      os.environ[env_vars.AGENT_PUBLIC_DIRECTORY_PATH] = kwargs['public_directory_path']

    if kwargs.get('response_fixtures_path'):
      os.environ[env_vars.AGENT_RESPONSE_FIXTURES_PATH] = kwargs['response_fixtures_path']

    # Observe config for changes
    Settings.instance().watch()

    if not os.getenv(env_vars.LOG_LEVEL):
        os.environ[env_vars.LOG_LEVEL] = kwargs['log_level']

    if 'api_url' in kwargs and kwargs['api_url']:
        os.environ[env_vars.API_URL] = kwargs['api_url']

    url = initialize_api(**kwargs)
    if 'headless' in kwargs and not kwargs['headless']:
        run_api(url)

    if 'proxyless' in kwargs and not kwargs['proxyless']:
        run_proxy(**kwargs)

@main.command(
  help="Mock request"
)
@click.option('-d', '--data', default='', help='HTTP POST data')
@ConditionalDecorator(lambda f: click.option('--remote-project-key', help='Use remote project for endpoint definitions.')(f), is_remote and is_local)
@click.option('--format', type=click.Choice([RAW_FORMAT]), help='Format response')
@click.option('-H', '--header', multiple=True, help='Pass custom header(s) to server')
@click.option('--lifecycle-hooks-path', help='Path to lifecycle hooks script.')
@click.option('-o', '--output', help='Write to file instead of stdout')
@ConditionalDecorator(lambda f: click.option('--project-key')(f), is_remote)
@click.option('--public-directory-path', help='Path to public files. Used for mocking requests.')
@click.option('--response-fixtures-path', help='Path to response fixtures yaml. Used for mocking requests.')
@click.option('-X', '--request', default='GET', help='Specify request command to use')
@click.option('--scenario-key')
@click.argument('url')
def mock(**kwargs):
  if kwargs.get('remote_project_key'):
    validate_project_key(kwargs['remote_project_key'])

  response = __replay(mode.MOCK, **kwargs)

  if response.status_code == custom_response_codes.NOT_FOUND:
    content = response.content
    print(f"Error: {decode(content)}")
    sys.exit(1)

  if kwargs['format'] == RAW_FORMAT:
    print_raw_response(response, kwargs['output'])
  else:
    content = response.content

    if not kwargs['output']:
      print(decode(content), end='')
    else:
      with open(kwargs['output'], 'w') as fp:
        fp.write(decode(content)) 

@main.command(
  help="Record request"
)
@click.option('-d', '--data', default='', help='HTTP POST data')
@click.option('--format', type=click.Choice([RAW_FORMAT]), help='Format response')
@click.option('-H', '--header', multiple=True, help='Pass custom header(s) to server')
@click.option('--lifecycle-hooks-path', help='Path to lifecycle hooks script.')
@click.option('-o', '--output', help='Write to file instead of stdout')
@ConditionalDecorator(lambda f: click.option('--project-key')(f), is_remote)
@click.option('-X', '--request', default='GET', help='Specify request command to use')
@click.option('--scenario-key')
@click.argument('url')
def record(**kwargs):
  response = __replay(mode.RECORD, **kwargs) 

  if kwargs['format'] == RAW_FORMAT:
    print_raw_response(response, kwargs['output'])
  else:
    content: bytes = response.raw.data

    if not kwargs['output']:
      try:
        print(content.decode(json.detect_encoding(content)), end='')
      except UnicodeDecodeError:
        print('Warning: Binary output can mess up your terminal.')
    else:
      with open(kwargs['output'], 'w') as fp:
        fp.write(content.decode(json.detect_encoding(content))) 

def __build_request_from_curl(**kwargs):
  headers = {}
  for header in kwargs['header']:
    toks = header.split(':')

    if len(toks) != 2:
      continue

    headers[toks[0].strip()] = toks[1].strip()

  return requests.Request(
    data=kwargs['data'],
    headers=headers,
    method=kwargs['request'],
    url=kwargs['url']
  )

def __replay(mode, **kwargs):
  if kwargs.get('scenario_key'):
    validate_scenario_key(kwargs['scenario_key'])

  request = __build_request_from_curl(**kwargs)

  context = ReplayContext.from_python_request(request)
  return replay_request(context, {
    **kwargs,
    'mode': mode,
  })



from io import BytesIO
import json
from requests import Response

from stoobly_agent.app.proxy.constants import custom_response_codes

class IgnoreComponentsResponseBuilder():
  __response = None

  def __init__(self):
    self.__response = Response()

  def build(self, ignored_components: list = []):
    content = json.dumps(ignored_components)
    self.__response.raw = BytesIO(content.encode()) 
    self.__response.status_code = custom_response_codes.IGNORE_COMPONENTS
    return self.__response


import copy
import pdb

from dataclasses import dataclass
from typing import List
from urllib.parse import urlencode
from urllib.parse import urlparse

from stoobly_agent.app.models.adapters.orm import (
  OrmRequestAdapterFactory,
  OrmResponseAdapterFactory,
)
from stoobly_agent.app.models.factories.resource.local_db.request_adapter import (
  LocalDBRequestAdapter,
)
from stoobly_agent.app.proxy.replay.body_parser_service import (
  decode_response,
  encode_response,
)
from stoobly_agent.app.proxy.test.context import TestContext
from stoobly_agent.app.proxy.test.helpers.endpoint_facade import EndpointFacade
from stoobly_agent.app.proxy.test.helpers.request_component_names_facade import (
  RequestComponentNamesFacade,
)
from stoobly_agent.app.proxy.test.matchers.contract import matches as contract_matches
from stoobly_agent.app.settings.constants import request_component
from stoobly_agent.config.constants.lifecycle_hooks import (
  ON_LENGTH_MATCH_ERROR,
  ON_PARAM_NAME_EXISTS_ERROR,
  ON_PARAM_NAME_MISSING_ERROR,
  ON_TYPE_MATCH_ERROR,
  ON_VALID_TYPE_ERROR,
)
from stoobly_agent.lib.api.interfaces.endpoints import EndpointShowResponse
from stoobly_agent.lib.api.interfaces.requests import RequestShowResponse
from stoobly_agent.lib.orm.request import Request

from .request_synchronize_handler import RequestSynchronizeHandler

@dataclass
class SynchronizeRequestService():
  local_db_request_adapter: LocalDBRequestAdapter

  def synchronize_request(self, request: Request, endpoint: EndpointShowResponse, lifecycle_hooks = {}):
    facade = EndpointFacade(None, -1).with_show_response(endpoint)
    mitmproxy_request = OrmRequestAdapterFactory(request).mitmproxy_request()

    # Query Params
    query_param_names_facade = facade.query_param_names
    handler = RequestSynchronizeHandler(request_component.QUERY_PARAM, lifecycle_hooks)
    query_params = mitmproxy_request.query
    result = self.synchronize_component(query_params, query_param_names_facade, handler)
    success = result[0]
    if success:
      self.__remove_bad_params(query_params)
      updated_url = self.__encode_query_params(request.url, request.path, query_params)
      params: RequestShowResponse = {}
      params['url'] = updated_url

      self.local_db_request_adapter.update(request.id, **params)

    # Body
    headers = mitmproxy_request.headers
    body_params = decode_response(mitmproxy_request.content, headers.get('content-type'))
    endpoint_body_params = endpoint.get('body_param_names')

    if body_params or endpoint_body_params:
      body_param_names_facade = facade.body_param_names
      handler = RequestSynchronizeHandler(request_component.BODY_PARAM, lifecycle_hooks)
      result = self.synchronize_component(body_params, body_param_names_facade, handler)

      success = result[0]
      if success:
        encoded_response = encode_response(body_params, headers.get('content-type'))
        params: RequestShowResponse = {}
        params['body'] = encoded_response

        self.local_db_request_adapter.update(request.id, **params)

    # Response
    python_response = OrmResponseAdapterFactory(request.response).python_response()
    headers = python_response.headers
    response_params = decode_response(python_response.content, headers.get('content-type'))
    response_param_names_facade = facade.response_param_names
    valid = self.validate_component(response_param_names_facade, response_params)[0]

    if not valid:
      self.__synchronize_response(endpoint)

  def synchronize_component(self, component, facade: RequestComponentNamesFacade, handler: RequestSynchronizeHandler):
    lifecycle_hooks = self.__lifecycle_hooks(handler)
    context = self.__context(lifecycle_hooks)

    return contract_matches(context, facade, component, strict=True)

  def validate_component(self, facade: RequestComponentNamesFacade, component):
    context = self.__context({})

    return contract_matches(context, facade, component)

  class TestContextMock(TestContext):
    def __init__(self, lifecycle_hooks: dict):
      self.__lifecycle_hooks = lifecycle_hooks

    @property
    def lifecycle_hooks(self):
      return self.__lifecycle_hooks

    @lifecycle_hooks.setter
    def lifecycle_hooks(self, v):
      self.__lifecycle_hooks = v

  def __context(self, lifecycle_hooks) -> TestContextMock:
    return self.TestContextMock(lifecycle_hooks)

  def __lifecycle_hooks(self, handler: RequestSynchronizeHandler):
    hooks = {}

    hooks[ON_LENGTH_MATCH_ERROR] = handler.handle_length_match_error
    hooks[ON_PARAM_NAME_EXISTS_ERROR] = handler.handle_param_name_exists
    hooks[ON_PARAM_NAME_MISSING_ERROR] = handler.handle_param_name_missing
    hooks[ON_TYPE_MATCH_ERROR] = handler.handle_type_match_error
    hooks[ON_VALID_TYPE_ERROR] = handler.handle_valid_type_error

    return hooks

  def __synchronize_response(self, endpoint: EndpointShowResponse):
    # For each request / scenario, replay and overwrite
    pass

  def __encode_query_params(self, original_url: str, path: str, query_params) -> str:
    query_tuples = []
    for key in query_params:
      for value in query_params.get_all(key):
        if not isinstance(value, list):
          query_tuples.append((key, value))
        else:
          for v in value:
            query_tuples.append((key, v))

    parsed_url = urlparse(original_url)
    url = parsed_url._replace(path=path, query=urlencode(query_tuples))
    return url.geturl()

  def __remove_bad_params(self, query_params) -> None:
    query_params_clone = copy.deepcopy(query_params)
    for key in query_params_clone:
      for value in query_params_clone.get_all(key):
        if value == 'None':
          del query_params[key]

  def __find_required_params(self, endpoint: EndpointShowResponse, param_name: str) -> List:
    result = []

    params = endpoint.get(param_name)
    if not params:
      return result

    for param in params:
      if param['is_required'] == True:
        result.append(param)

    return result




from typing import Literal

HEADER = 'Header'
BODY_PARAM = 'Body Param'
QUERY_PARAM = 'Query Param'
RESPONSE_HEADER = 'Response Header'
RESPONSE_PARAM = 'Response Param'

REQUEST_COMPONENTS = [BODY_PARAM, HEADER, QUERY_PARAM]
RESPONSE_COMPONENTS = [RESPONSE_HEADER, RESPONSE_PARAM]

RequestComponent = Literal[BODY_PARAM, HEADER, QUERY_PARAM, RESPONSE_HEADER, RESPONSE_PARAM]


class ConditionalDecorator():

  def __init__(self, decorator, condition):
    self.__decorator = decorator
    self.__condition = condition

  def __call__(self, f):
    if not self.__condition:
      return f
    
    return self.__decorator(f)


import json
import pdb
import pytest

from click.testing import CliRunner
from pathlib import Path

from stoobly_agent.test.test_helper import reset

from stoobly_agent.app.models.adapters.raw_http_request_adapter import RawHttpRequestAdapter
from stoobly_agent.app.models.types import OPENAPI_FORMAT
from stoobly_agent.app.settings import Settings
from stoobly_agent.cli import endpoint
from stoobly_agent.lib.orm.request import Request

from stoobly_agent.app.models.factories.resource.local_db.helpers.request_builder import RequestBuilder

@pytest.fixture(scope='module')
def runner():
    return CliRunner()

@pytest.fixture(scope='module')
def mock_data_directory_path() -> str:
  return Path(__file__).parent.parent.parent.parent / 'mock_data'

@pytest.fixture(scope='module')
def petstore_expanded_file_path(mock_data_directory_path) -> str:
  path = mock_data_directory_path / "petstore-expanded.yaml"
  return str(path)

@pytest.mark.openapi
class TestApply():
    @pytest.fixture(scope='class', autouse=True)
    def settings(self):
        return reset()

    @pytest.fixture(scope='class')
    def request_body_one(self):
        return {
            'name': 'pet1',
            'tag': 'best',
        }

    @pytest.fixture(scope='class', autouse=True)
    def created_request_extra_body_param(self, settings: Settings, request_body_one):
        body = {
            **request_body_one,
            'delete': 1,
        }

        status = RequestBuilder(
            method='POST',
            request_body=body,
            request_headers={
                'content-type': 'application/json'
            },
            response_body=body,
            response_headers={
                'content-type': 'application/json'
            },
            status_code=200,
            url='https://petstore.swagger.io/v2/pets',
        ).with_settings(settings).build()[1]
        assert status == 200

        return Request.last()

    @pytest.fixture(scope='class')
    def request_body_two(self):
        return {
            'name': 'pet1',
        }

    @pytest.fixture(scope='class', autouse=True)
    def created_request_missing_body_param(self, settings: Settings):
        body = {
          'tag': 'best',
        }

        status = RequestBuilder(
            method='POST',
            request_body=body,
            request_headers={
                'content-type': 'application/json'
            },
            response_body=body,
            response_headers={
                'content-type': 'application/json'
            },
            status_code=200,
            url='https://petstore.swagger.io/v2/pets',
        ).with_settings(settings).build()[1]
        assert status == 200

        return Request.last()

    class TestWhenApplyingPestoreExpanded():

        @pytest.fixture(scope='class', autouse=True)
        def import_result(self, runner: CliRunner, petstore_expanded_file_path: str):
            record_result = runner.invoke(endpoint, ['apply', '--source-format', OPENAPI_FORMAT, '--source-path', petstore_expanded_file_path])
            assert record_result.exit_code == 0
            return record_result

        def test_it_deletes(self, created_request_extra_body_param: Request, request_body_one):
            _created_request = Request.find(created_request_extra_body_param.id)
            python_request = RawHttpRequestAdapter(_created_request.raw).to_request()
            assert python_request.data == json.dumps(request_body_one).encode()

        def test_it_adds(self, created_request_missing_body_param: Request):
            _created_request = Request.find(created_request_missing_body_param.id)
            python_request = RawHttpRequestAdapter(_created_request.raw).to_request()

            assert python_request.data == json.dumps({
                'tag': 'best',
                'name': '',
            }).encode()




import pdb
import requests

from typing import Tuple

from stoobly_agent.app.models.types import RequestCreateParams, RequestShowParams
from stoobly_agent.lib.api.requests_resource import RequestsResource
from stoobly_agent.lib.api.interfaces import RequestShowResponse, RequestsIndexResponse
from stoobly_agent.lib.orm.types.request_columns import RequestColumns

class StooblyRequestAdapter():

  def __init__(self, __api: RequestsResource):
    self.__api = __api

  def create(self, **params: RequestCreateParams):
    body_params = { **params }
    joined_request = body_params['joined_request']

    del body_params['flow']
    del body_params['joined_request']

    res = self.__api.create(**{ 'importer': 'gor', 'requests': joined_request.build(), **body_params })
    res.raise_for_status()  
    return res.json(), res.status_code

  def response(self, **query_params: RequestColumns):
    return self.__api.response(**query_params)

  def show(self, request_id: str, **params: RequestShowParams) -> Tuple[RequestShowResponse, int]:
    query_params = { **params } 

    res = self.__api.show(request_id, **query_params)
    res.raise_for_status()  
    return res.json(), res.status_code

  def index(self, **query_params) -> Tuple[RequestsIndexResponse, int]:
    res = self.__api.index(**query_params)
    res.raise_for_status()  
    return res.json(), res.status_code



import pdb

from stoobly_agent.app.models.adapters.joined_request_adapter import JoinedRequestAdapter
from stoobly_agent.app.models.adapters.raw_http_response_adapter import RawHttpResponseAdapter

class PythonResponseAdapter():

  def __init__(self, joined_request: bytes):
    self.__request = joined_request

  def adapt(self):
    if not self.__request:
      return None

    adapter = JoinedRequestAdapter(self.__request)
    response_string = adapter.build_response_string().get()
    return RawHttpResponseAdapter(response_string).to_response()


import os
import pdb

from .app import App
from .app_command import AppCommand
from .service_config import ServiceConfig


class ServiceCommand(AppCommand):

  def __init__(self, app: App, **kwargs):
    super().__init__(app)
    self.__service_name = kwargs.get('service_name')

    self.__config = ServiceConfig(self.service_path, **kwargs)

  @property
  def service_config(self):
    return self.__config

  @property
  def service_config_path(self):
    return self.__config.path

  @property
  def service_name(self):
    return self.__service_name

  @property
  def service_exists(self):
    return os.path.exists(self.service_path)

  @property
  def service_path(self):
    return os.path.join(
      self.data_dir_path,
      self.service_relative_path
    )

  @property
  def service_relative_path(self):
    return os.path.join(
      self.namespace,
      self.service_name,
    )

  def config(self, _c: dict):
    _config = self.app_config.read()
    _config.update(self.service_config.read())
    _config.update(_c)
    return _config




import os

from stoobly_agent.app.settings import Settings
from stoobly_agent.config.constants import env_vars
from stoobly_agent.lib.api.keys.project_key import InvalidProjectKey, ProjectKey

def remote(settings: Settings):
  return settings.cli.features.remote or not not os.environ.get(env_vars.FEATURE_REMOTE)

def local(settings: Settings):
  is_local = True

  if remote(settings):
    project_key = settings.proxy.intercept.project_key

    try:
      return ProjectKey(project_key).is_local
    except InvalidProjectKey:
      pass

  return is_local


import time

from mitmproxy.http import HTTPFlow as MitmproxyHTTPFlow
from requests import Response

from stoobly_agent.app.proxy.intercept_settings import InterceptSettings

from ..context import InterceptContext

class MockContext(InterceptContext):
  def __init__(self, flow: MitmproxyHTTPFlow, intercept_settings: InterceptSettings):
    super().__init__(flow, intercept_settings)

    self.__start_time = time.time()
    self.__end_time = None
    self.__response = None

  @property
  def end_time(self):
    return self.__end_time

  @property
  def response(self):
    return self.__response

  @property
  def start_time(self):
    return self.__start_time

  def with_response(self, response: Response):
    self.__response = response
    self.__end_time = time.time()
    return self


import json
import pdb

from typing import Union
from diff_match_patch import diff_match_patch

green = '\x1b[38;5;16;48;5;2m'
red = '\x1b[38;5;16;48;5;1m'
endgreen = '\x1b[0m'
endred = '\x1b[0m'

error = 'Error: Cannot print diff for binary output'

def diff(a: Union[bytes, str], b: Union[str, bytes], timeout = 30) -> str:

  if not __string_like(a):
    a = json.dumps(a, indent=2)
  else:
    try:
      a = a.decode() if isinstance(a, bytes) else a
    except UnicodeDecodeError:
      return error

  if not __string_like(b):
    b = json.dumps(b, indent=2)
  else:
    try:
      b = b.decode() if isinstance(b, bytes) else b
    except UnicodeDecodeError:
      return error

  output = []
  dmp = diff_match_patch()
  for delta in dmp.diff_lineMode(a, b, timeout):
    opcode = delta[0]
    text = delta[1]

    if opcode == 0:
      output.append(text)
    elif opcode == 1:
      length = len(text)

      i = 0
      j = 0
      while i < length:
        if text[i] == "\n":
          output.append(f"{green}{text[j:i]}{endgreen}\n")
          j = i + 1

        i += 1

      if i != j:
        output.append(f"{green}{text[j:i]}{endgreen}")
    elif opcode == -1:
      length = len(text)

      i = 0
      j = 0
      while i < length:
        if text[i] == "\n":
          output.append(f"{red}{text[j:i]}{endred}\n")
          j = i + 1

        i += 1

      if i != j:
        output.append(f"{red}{text[j:i]}{endred}")

  return ''.join(output)

def __string_like(s):
  return isinstance(s, str) or isinstance(s, bytes)


import click
import pdb
import sys

from stoobly_agent.config.data_dir import DataDir

from .helpers.certificate_authority import CertificateAuthority

@click.group(
    epilog="Run 'stoobly-agent COMMAND --help' for more information on a command.",
    help="Manage CA certificate"
)
@click.pass_context
def ca_cert(ctx):
    pass

@ca_cert.command()
@click.option('--ca-certs-dir-path', default=DataDir.instance().ca_certs_dir_path, help='Path to ca certs directory used to sign SSL certs.')
@click.option('--certs-dir-path', default=DataDir.instance().certs_dir_path, help='Output directory.')
@click.argument('hostname')
def mkcert(**kwargs):
    ca_certs_dir_path = kwargs['ca_certs_dir_path']

    installer = CertificateAuthority(ca_certs_dir_path)

    try:
        installer.sign(kwargs['hostname'], kwargs['certs_dir_path'])
    except Exception as e:
        print(e, file=sys.stderr)
        sys.exit(1)

@ca_cert.command()
@click.option('--ca-certs-dir-path', default=DataDir.instance().ca_certs_dir_path, help='Path to ca certs directory.')
def install(**kwargs):
    ca_certs_dir_path = kwargs['ca_certs_dir_path']
    installer = CertificateAuthority(ca_certs_dir_path)

    try:
        installer.install()
    except Exception as e:
        print(e, file=sys.stderr)
        sys.exit(1)

@ca_cert.command()
def uninstall():
    print("Not yet implemented. Stay tuned!")


from io import BytesIO
from requests import Response

from stoobly_agent.app.proxy.constants import custom_response_codes

class CustomNotFoundResponseBuilder():
  __response = None

  def __init__(self):
    self.__response = Response()

  def build(self):
    self.__response.status_code = custom_response_codes.NOT_FOUND
    self.__response.raw = BytesIO('Request not found. To troubleshoot see https://docs.stoobly.com/guides/how-to-mock-apis/troubleshooting'.encode()) 
    self.__response.headers = {
      'Access-Control-Allow-Origin': '*',
      'Access-Control-Allow-Methods': 'GET, OPTIONS, POST, PATCH, PUT, DELETE',
      'Content-Type': 'text/plain',
    }
    return self.__response


import pdb
import pytest

from stoobly_agent.app.proxy.test.helpers.test_results_builder import DownloadTestData, TestResultsBuilder as Builder
from stoobly_agent.config.constants import test_strategy

class TestBuilder():

  class TestWhenValid():
    @pytest.fixture(scope='class')
    def test_results(self) -> DownloadTestData:
      return {
        'expected_response': b'expected',
        'log': 'test',
        'passed': True,
        'request_id': 1,
        'received_response': b'received',
        'skipped': False,
        'status': 200,
        'strategy': test_strategy.FUZZY,
      }

    @pytest.fixture(scope='class')
    def builder(self, test_results) -> Builder:
      _builder = Builder(**test_results)
      builder = Builder()
      builder.deserialize(_builder.serialize())
      return builder

    def test_it_has_expected_response(self, builder: Builder, test_results: DownloadTestData):
      assert builder.expected_response == test_results['expected_response']

    def test_it_has_log(self, builder: Builder, test_results: DownloadTestData):
      assert builder.log == test_results['log']

    def test_it_has_passed(self, builder: Builder, test_results: DownloadTestData):
      assert builder.passed == test_results['passed']

    def test_it_has_request_id(self, builder: Builder, test_results: DownloadTestData):
      assert builder.request_id == test_results['request_id']

    def test_it_has_received_response(self, builder: Builder, test_results: DownloadTestData):
      assert builder.received_response == test_results['received_response']

    def test_it_has_skipped(self, builder: Builder, test_results: DownloadTestData):
      assert builder.skipped == test_results['skipped']

    def test_it_has_status(self, builder: Builder, test_results: DownloadTestData):
      assert builder.status == test_results['status']

    def test_it_has_strategy(self, builder: Builder, test_results: DownloadTestData):
      assert builder.strategy == test_results['strategy']

  class TestWhenEmptyResponse():
    @pytest.fixture(scope='class')
    def test_results(self) -> DownloadTestData:
      return {
        'expected_response': '',
        'received_response': '',
      }

    @pytest.fixture(scope='class')
    def builder(self, test_results) -> Builder:
      _builder = Builder(**test_results)
      builder = Builder()
      builder.deserialize(_builder.serialize())
      return builder

    def test_it_has_expected_response(self, builder: Builder, test_results: DownloadTestData):
      assert builder.expected_response == test_results['expected_response']

    def test_it_has_received_response(self, builder: Builder, test_results: DownloadTestData):
      assert builder.received_response == test_results['received_response']

  class TestWhenTraversibleResponse():
    @pytest.fixture(scope='class')
    def test_results(self) -> DownloadTestData:
      return {
        'expected_response': {},
        'received_response': {},
      }

    @pytest.fixture(scope='class')
    def builder(self, test_results) -> Builder:
      _builder = Builder(**test_results)
      builder = Builder()
      builder.deserialize(_builder.serialize())
      return builder

    def test_it_has_expected_response(self, builder: Builder, test_results: DownloadTestData):
      assert builder.expected_response == test_results['expected_response']

    def test_it_has_received_response(self, builder: Builder, test_results: DownloadTestData):
      assert builder.received_response == test_results['received_response']


import os
import pdb

from mitmproxy.http import HTTPFlow as MitmproxyHTTPFlow
from mitmproxy.http import Headers, Request as MitmproxyRequest

from stoobly_agent.app.proxy.context import InterceptContext
from stoobly_agent.app.proxy.handle_mock_service import handle_request_mock, handle_response_mock
from stoobly_agent.app.proxy.handle_replay_service import handle_request_replay, handle_response_replay
from stoobly_agent.app.proxy.handle_record_service import handle_response_record
from stoobly_agent.app.proxy.handle_test_service import handle_request_test, handle_response_test
from stoobly_agent.app.proxy.intercept_settings import InterceptSettings
from stoobly_agent.app.proxy.mock.context import MockContext
from stoobly_agent.app.proxy.replay.context import ReplayContext
from stoobly_agent.app.proxy.record.context import RecordContext
from stoobly_agent.app.proxy.utils.response_handler import bad_request
from stoobly_agent.app.settings import Settings
from stoobly_agent.config.constants import lifecycle_hooks, mode
from stoobly_agent.lib.logger import Logger

# Disable proxy settings in urllib
os.environ['no_proxy'] = '*'

LOG_ID = 'Intercept'

def request(flow: MitmproxyHTTPFlow):
    request: MitmproxyRequest = flow.request

    __patch_cookie(request)

    intercept_settings = InterceptSettings(Settings.instance(), request)

    if not intercept_settings.active:
        return

    __disable_web_cache(request)
    __intercept_hook(lifecycle_hooks.BEFORE_REQUEST, flow, intercept_settings)

    active_mode = intercept_settings.mode
    Logger.instance(LOG_ID).debug(f"ProxyMode: {active_mode}")

    if active_mode == mode.MOCK:
        context = MockContext(flow, intercept_settings)
        handle_request_mock(context)
    elif active_mode == mode.RECORD:
        pass
    elif active_mode == mode.REPLAY:
        context = ReplayContext(flow, intercept_settings)
        handle_request_replay(context)
    elif active_mode == mode.TEST:
        context = ReplayContext(flow, intercept_settings)
        handle_request_test(context)
    else:
        if active_mode != mode.NONE:
            bad_request(
                flow,
                "Valid env MODES: %s, Got: %s" % ([mode.MOCK, mode.RECORD, mode.REPLAY, mode.TEST], active_mode)
            )

def response(flow: MitmproxyHTTPFlow):
    request: MitmproxyRequest = flow.request

    intercept_settings = InterceptSettings(Settings.instance(), request)
    intercept_settings.for_response()

    if not intercept_settings.active:
        return

    active_mode = intercept_settings.mode

    if active_mode == mode.MOCK:
        context = MockContext(flow, intercept_settings)
        handle_response_mock(context)
    elif active_mode == mode.RECORD:
        context = RecordContext(flow, intercept_settings)
        handle_response_record(context)
    elif active_mode == mode.REPLAY:
        context = ReplayContext(flow, intercept_settings)
        handle_response_replay(context)
    elif active_mode == mode.TEST:
        context = ReplayContext(flow, intercept_settings)
        handle_response_test(context)

    __intercept_hook(lifecycle_hooks.BEFORE_RESPONSE, flow, intercept_settings)

### PRIVATE

# Prevent 304 status
# Because this header will get recorded, should add during mocking as well in the case where headers are used for matching
def __disable_web_cache(request: MitmproxyRequest) -> None:
    request.headers['Cache-Control'] = 'no-cache, no-store'

# Fix issue where multi-value cookies become comma separated
def __patch_cookie(request: MitmproxyRequest):
    header_name = 'cookie'

    if len(request.headers.get_all(header_name)) > 1:
        __combine_header(request.headers, header_name, '; ')
        
def __combine_header(headers: Headers, header_name: str, delimitter: str):
    values = headers.get_all(header_name)
    headers[header_name] = delimitter.join(values)

def __intercept_hook(hook: str, flow: MitmproxyHTTPFlow, intercept_settings: InterceptSettings):
    lifecycle_hooks_module = intercept_settings.lifecycle_hooks

    if hook in lifecycle_hooks_module:
        lifecycle_hooks_module[hook](InterceptContext(flow, intercept_settings))


import pdb
import pytest

from click.testing import CliRunner
from pathlib import Path

from stoobly_agent.cli import mock, record, request
from stoobly_agent.lib.orm.request import Request
from stoobly_agent.test.test_helper import DETERMINISTIC_GET_REQUEST_URL, reset

@pytest.fixture(scope='module')
def runner():
  return CliRunner()

class TestLifecycleHooks():
  @pytest.fixture(scope='class', autouse=True)
  def settings(self):
    return reset()

  @pytest.fixture(scope='class')
  def lifecycle_hooks_path(self):
    return str(Path(__file__).parent.parent / 'mock_data' / 'lifecycle_hooks.py')

  @pytest.fixture(scope='class')
  def record_result(self, runner: CliRunner, lifecycle_hooks_path: str):
    record_result = runner.invoke(record, [
      '--lifecycle-hooks-path', lifecycle_hooks_path, '--output', '/dev/null', DETERMINISTIC_GET_REQUEST_URL
    ])
    assert record_result.exit_code == 0
    return record_result

  @pytest.fixture(scope='class')
  def recorded_request(self):
    return Request.last()

  @pytest.fixture(scope='class')
  def mock_result(self, runner: CliRunner, lifecycle_hooks_path: str):
    mock_result = runner.invoke(mock, [
      '--lifecycle-hooks-path', lifecycle_hooks_path, '--output', '/dev/null', DETERMINISTIC_GET_REQUEST_URL
    ])
    assert mock_result.exit_code == 0
    return mock_result

  @pytest.fixture(scope='class')
  def test_result(self, runner: CliRunner, lifecycle_hooks_path: str, recorded_request: Request):
    test_result = runner.invoke(request, [
      'test', '--format', 'json', '--lifecycle-hooks-path', lifecycle_hooks_path, recorded_request.key() 
    ])
    assert test_result.exit_code == 0
    return test_result

  def test_calls_record_hooks(self, record_result):
    expected_stdout = ['before_request', 'before_record', 'after_record', 'before_response']
    assert record_result.stdout.strip() == "\n".join(expected_stdout)

  def test_calls_mock_hooks(self, mock_result):
    expected_stdout = ['before_request', 'before_mock', 'after_mock', 'before_response']
    assert mock_result.stdout.strip() == "\n".join(expected_stdout)

  def test_calls_test_hooks(self, test_result):
     expected_stdout = [
       'before_request', 'before_replay', 'after_replay', 'before_mock', 'after_mock', 'before_test', 'after_test', 'before_response'
     ]
     stdout = test_result.stdout
     lifecycle_hooks_output = stdout.split('{')[0]
     assert lifecycle_hooks_output.strip() == "\n".join(expected_stdout)


import pdb
import pytest

from stoobly_agent.test.test_helper import reset

from stoobly_agent.config.data_dir import DataDir
from stoobly_agent.app.cli.helpers.certificate_authority import CertificateAuthority

@pytest.fixture(autouse=True, scope='module')
def settings():
  return reset()

class TestCertificateAuthority():

  @pytest.fixture(autouse=True, scope='class')
  def certificate_authority(self):
    return CertificateAuthority(DataDir.instance().certs_dir_path)

  class TestWhenCertDirConfigured():

    def test_defaults_to_not_installed(self, certificate_authority: CertificateAuthority):
      assert not certificate_authority.certs_generated

    def test_generates(self, certificate_authority: CertificateAuthority):
      certificate_authority.generate_certs()
      assert certificate_authority.certs_generated

  class TestWhenCertDirNotConfigured():

    @pytest.fixture(autouse=True, scope='class')
    def certificate_authority(self):
      return CertificateAuthority()

    def test_defaults_to_mitmproxy_confdir(self, certificate_authority: CertificateAuthority):
      assert certificate_authority.certs_dir == DataDir.instance().ca_certs_dir_path

  class TestSign():

    class TestWhenExistentHostname():

      @pytest.fixture(scope='class')
      def hostname(self):
        return 'www.google.com'

      @pytest.fixture(autouse=True, scope='class')
      def sign(self, certificate_authority: CertificateAuthority, hostname: str):
        certificate_authority.sign(hostname)

      def test_signed(self, certificate_authority: CertificateAuthority, hostname: str):
        assert certificate_authority.signed(hostname)

    class TestWhenNonExistentHostname():

      @pytest.fixture(scope='class')
      def hostname(self):
        return 'test.stoobly.com'

      @pytest.fixture(autouse=True, scope='class')
      def sign(self, certificate_authority: CertificateAuthority, hostname: str):
        certificate_authority.sign(hostname)

      def test_signed(self, certificate_authority: CertificateAuthority, hostname: str):
        assert certificate_authority.signed(hostname)

    class TestWhenSigned():

      @pytest.fixture(scope='class')
      def hostname(self):
        return 'www.google.com'

      @pytest.fixture(autouse=True, scope='class')
      def sign(self, certificate_authority: CertificateAuthority, hostname: str):
        certificate_authority.sign(hostname)

      def test_it_does_not_sign(self, certificate_authority: CertificateAuthority, hostname: str):
        assert not certificate_authority.sign(hostname)


import pdb

from datetime import datetime

from stoobly_agent.app.api.simple_http_request_handler import SimpleHTTPRequestHandler
from stoobly_agent.app.models.adapters.orm import JoinedRequestStringAdapter
from stoobly_agent.app.models.scenario_model import ScenarioModel
from stoobly_agent.app.settings import Settings
from stoobly_agent.lib.orm.scenario import Scenario
from stoobly_agent.lib.utils.decode import decode

class ScenariosController:
    _instance = None

    def __init__(self):
        if self._instance:
            raise RuntimeError('Call instance() instead')
        else:
            self.data = {}

    @classmethod
    def instance(cls):
        if cls._instance is None:
            cls._instance = cls()

        return cls._instance

    # POST /scenarios
    def create(self, context: SimpleHTTPRequestHandler):
        body_params = context.params
        if not context.required_params(body_params, ['name']):
            return

        description = decode(body_params['description'] if 'description' in body_params else '')
        name = decode(body_params['name'])
        priority = decode(body_params['priority'] if 'priority' in body_params else '0')

        scenario_model = self.__scenario_model(context)
        scenario, status = scenario_model.create(**{
            'name':name, 
            'description': description,
            'priority': priority,
        })

        if context.filter_response(scenario, status):
            return

        context.render(
            json = scenario,
            status = 200
        )

    # GET /scenarios/:id
    def get(self, context: SimpleHTTPRequestHandler):
        context.parse_path_params({
            'id': 1
        })

        scenario_model = self.__scenario_model(context)
        scenario, status = scenario_model.show(context.params.get('id'))

        if context.filter_response(scenario, status):
            return
        
        context.render(
            json = scenario,
            status = 200
        )

    # GET /scenarios
    def index(self, context: SimpleHTTPRequestHandler):
        scenario_model = self.__scenario_model(context)
        scenarios, status = scenario_model.index(**context.params)

        if context.filter_response(scenarios, status):
            return

        context.render(
            json = scenarios,
            status = 200
        )

    # PUT /scenarios/:id
    def update(self, context: SimpleHTTPRequestHandler):
        context.parse_path_params({
            'id': 1
        })

        scenario_id = context.params.get('id')

        scenario_model = self.__scenario_model(context)
        scenario, status = scenario_model.update(scenario_id, **context.params.get('scenario'))

        if context.filter_response(scenario, status):
            return
            
        context.render(
            json = scenario,
            status = 200
        )

    # DELETE /scenarios/:id
    def destroy(self, context: SimpleHTTPRequestHandler):
        context.parse_path_params({
            'id': 1
        })

        scenario_id = context.params.get('id')

        scenario_model = self.__scenario_model(context)
        scenario, status = scenario_model.destroy(scenario_id)

        if context.filter_response(scenario, status):
            return

        context.render(
            plain = '',
            status = 200
        )

    def download(self, context: SimpleHTTPRequestHandler):
        context.parse_path_params({
            'id': 1
        })
        scenario_id = int(context.params.get('id'))
        format = context.params.get('format')
        scenario = Scenario.find(scenario_id)

        if not scenario:
            return context.not_found()

        if format == 'gor':
            filename = f"SCENARIO-{int(datetime.now().timestamp())}.gor"
            requests = scenario.requests

            content = ''
            for request in requests:
                content = JoinedRequestStringAdapter(request).adapt(content)

            context.render(
                download = content,
                filename = filename,
                status = 200
            )
        else:
            return context.bad_request('Invalid format')

    def __scenario_model(self, context: SimpleHTTPRequestHandler):
        access_token = context.headers.get('access-token')
        scenario_model = ScenarioModel(Settings.instance(), access_token=access_token)
        return scenario_model


import pdb

from typing import TypedDict

from stoobly_agent.lib.logger import DEBUG

from .app import App
from .constants import WORKFLOW_CONTAINER_TEMPLATE
from .workflow_command import WorkflowCommand

class WorkflowLogCommand(TypedDict):
  container: str
  follow: bool
  namespace: str

class WorkflowLogCommand(WorkflowCommand):
  def __init__(self, app: App, **kwargs):
    super().__init__(app, **kwargs)

  def build(self, **options: WorkflowLogCommand):
    commands = []
    containers = self.containers
    allowed_containers = list(
      map(
        lambda container: WORKFLOW_CONTAINER_TEMPLATE.format(
          container=container, service_name=self.service_name
        ), options.get('containers') or []
      )
    )

    for index, container in enumerate(containers):
      if container not in allowed_containers:
        continue

      container_name = self.container_name(container, options.get('namespace'))
      commands.append(f"echo \"=== Logging {container_name}\"")
      if options.get('follow') and index == len(containers) - 1:
        command = ['docker', 'logs', '--follow', container_name]
      else:
        command = ['docker', 'logs', container_name]
      commands.append(' '.join(command))

    return commands

  def container_name(self, container, namespace=None):
    return f"{namespace or self.workflow_name}-{container}-1"


from stoobly_agent.config.data_dir import DataDir


class ServiceDockerCompose():
  def __init__(self, app_dir_path, target_workflow_name, service_name, hostname):
    self.service_name = service_name
    self.hostname = hostname
    self.container_name = f"{target_workflow_name}-{service_name}-1"
    self.proxy_container_name = f"{target_workflow_name}-{service_name}.proxy-1"
    self.init_container_name = f"{target_workflow_name}-{service_name}.init-1"
    self.configure_container_name = f"{target_workflow_name}-{service_name}.configure-1"

    data_dir_path = DataDir.instance(app_dir_path).path
    self.docker_compose_path = f"{data_dir_path}/docker/{service_name}/{target_workflow_name}/docker-compose.yml"
    self.init_script_path = f"{data_dir_path}/docker/{service_name}/{target_workflow_name}/bin/init"



import os
import pdb

from stoobly_orator.migrations import Migrator, DatabaseMigrationRepository

from stoobly_agent.config.data_dir import DataDir
from stoobly_agent.config.source_dir import SourceDir
from stoobly_agent.lib.orm import ORM

def migrate(version, pretend = False):
  # If current db version is same as version, return
  if os.path.exists(DataDir.instance().db_version_path):
    with open(DataDir.instance().db_version_path, 'r') as fp:
      _version = fp.read().strip()

      if _version == version:
        return

  migrator = __build_migrator()
  migrations_path = __build_migrations_path()
  migrator.run(migrations_path, pretend)

  with open(DataDir.instance().db_version_path, 'w') as fp:
    fp.write(version)

def rollback(pretend = False):
  migrator = __build_migrator()
  migrations_path = __build_migrations_path()
  migrator.rollback(migrations_path, pretend)

  os.remove(DataDir.instance().db_version_path)

def __build_migrator() -> Migrator:
  db = ORM.instance().db

  # Table to store migrations history
  repository = DatabaseMigrationRepository(db, 'migrations')
  migrator = Migrator(repository, db)

  if not migrator.repository_exists():
      repository.create_repository()

  return migrator

def __build_migrations_path():
  return os.path.join(SourceDir.instance().db_migrations_dir_path)


from stoobly_agent.app.settings import RemoteSettings
from stoobly_agent.lib.orm.request import Request

from .local_db.header_adapter import LocalDBHeaderAdapter

class HeaderResourceFactory():

  def __init__(self, settings: RemoteSettings):
    self.__remote_settings = settings

  def local_db(self) -> LocalDBHeaderAdapter:
    return LocalDBHeaderAdapter(Request)  


import pdb
import pytest
import time

from click.testing import CliRunner

from stoobly_agent.app.models.adapters.joined_request_adapter import JoinedRequestAdapter
from stoobly_agent.app.models.adapters.raw_http_response_adapter import RawHttpResponseAdapter
from stoobly_agent.app.models.factories.resource.local_db.helpers.log import Log
from stoobly_agent.app.models.factories.resource.local_db.helpers.log_event import LogEvent
from stoobly_agent.cli import record, request, snapshot
from stoobly_agent.lib.orm.request import Request
from stoobly_agent.test.test_helper import DETERMINISTIC_GET_REQUEST_URL, reset

@pytest.fixture(scope='module')
def runner():
  return CliRunner()

class TestUpdate():
  @pytest.fixture(scope='class', autouse=True)
  def settings(self):
    return reset()

  class TestEventColumns():

    @pytest.fixture(scope='class')
    def recorded_request(self, runner: CliRunner):
      record_result = runner.invoke(record, [DETERMINISTIC_GET_REQUEST_URL])
      assert record_result.exit_code == 0
      return Request.last()

    @pytest.fixture(autouse=True, scope='class')
    def put_event(self, runner: CliRunner, recorded_request: Request):
      snapshot_result = runner.invoke(request, ['snapshot', recorded_request.key()])
      assert snapshot_result.exit_code == 0

      log = Log()
      events = log.events
      return events[len(events) - 1]

    @pytest.fixture(autouse=True, scope='class')
    def updated_event(self, runner: CliRunner, put_event: LogEvent):
      time.sleep(1) # Simulate update at different time
      apply_result = runner.invoke(snapshot, ['update', put_event.uuid])
      assert apply_result.exit_code == 0
      log = Log()
      events = log.events
      assert len(events) == 2
      return events[1]

    def test_it_has_different_uuid(self, updated_event: LogEvent, put_event: LogEvent):
      assert updated_event.uuid != put_event.uuid 

    def test_it_has_same_action(self, updated_event: LogEvent, put_event: LogEvent):
      assert updated_event.action == put_event.action

    def test_it_has_same_resource(self, updated_event: LogEvent, put_event: LogEvent):
      assert updated_event.resource == put_event.resource

    def test_it_has_same_resource_uuid(self, updated_event: LogEvent, put_event: LogEvent):
      assert updated_event.resource_uuid == put_event.resource_uuid

    def test_it_has_different_created_at(self, updated_event: LogEvent, put_event: LogEvent):
      assert updated_event.created_at != put_event.created_at

  class TestVerifyOption():

    @pytest.fixture(scope='class')
    def recorded_request(self, runner: CliRunner):
      record_result = runner.invoke(record, [DETERMINISTIC_GET_REQUEST_URL])
      assert record_result.exit_code == 0
      return Request.last()

    @pytest.fixture(autouse=True, scope='class')
    def put_event(self, runner: CliRunner, recorded_request: Request):
      snapshot_result = runner.invoke(request, ['snapshot', recorded_request.key()])
      assert snapshot_result.exit_code == 0

      log = Log()
      events = log.events
      return events[len(events) - 1]

    @pytest.fixture(autouse=True, scope='class')
    def snapshot_contents(self, put_event: LogEvent):
      snapshot = put_event.snapshot()
      data_path = snapshot.path
      with open(data_path, 'ab') as fp:
        fp.write(b'test')

      with open(data_path, 'rb') as fp:
        return fp.read()

    @pytest.fixture(autouse=True, scope='class')
    def updated_event(self, runner: CliRunner, put_event: LogEvent):
      time.sleep(1) # Simulate update at different time

      apply_result = runner.invoke(snapshot, ['update', '--verify', put_event.uuid])
      assert apply_result.exit_code == 0
      log = Log()
      events = log.events
      assert len(events) == 2
      return events[1]

    def test_it_updates_snapshot_contents(self, snapshot_contents: str, updated_event: LogEvent):
      snapshot = updated_event.snapshot()
      data_path = snapshot.path

      with open(data_path, 'rb') as fp:
        contents = fp.read()
        new_response = self.__to_python_response(contents)
        content_length = str(len(new_response.content)).encode()

        assert content_length in contents
        assert content_length not in snapshot_contents

    def __to_python_response(self, raw_request):
      adapter = JoinedRequestAdapter(raw_request)
      response_string = adapter.build_response_string().get()
      return RawHttpResponseAdapter(response_string).to_response()


import pdb

from stoobly_agent.lib.orm.request import Request

from .mitmproxy_adapter import MitmproxyRequestAdapter

class OrmRequestAdapterFactory():

  def __init__(self, request: Request):
    self.__request = request

  def mitmproxy_request(self):
    return MitmproxyRequestAdapter(self.__request).adapt()


import pdb
from typing import Collection

from stoobly_agent.config.constants import alias_resolve_strategy
from stoobly_agent.lib.logger import Logger, bcolors
from stoobly_agent.lib.orm.trace import Trace
from stoobly_agent.lib.orm.trace_alias import TraceAlias

LOG_ID = 'Alias'

class AliasResolver():

  def __init__(self, trace: Trace, strategy: alias_resolve_strategy.AliasResolveStrategy):
    self.__trace = trace
    self.__strategy  = strategy

  @property
  def strategy(self):
    return self.__strategy

  @strategy.setter
  def strategy(self, v: alias_resolve_strategy.AliasResolveStrategy):
    self.__strategy = v

  def assign_alias(self, trace_alias: TraceAlias, value):
    if not trace_alias:
      return
      
    if trace_alias.assigned_to == None:
      trace_alias.assigned_to = value
      trace_alias.save()

      Logger.instance(LOG_ID).info(f"{bcolors.OKBLUE}Assigned{bcolors.ENDC} {trace_alias.name}: {value} -> {trace_alias.value}")

  def resolve_alias(self, alias_name: str, value: str):
    trace_alias = self.__resolve_alias(alias_name, value)
    return trace_alias

  def create_alias(self, alias_name, value, trace_request = None):
    columns = {
      'name': alias_name,
      'value': self.__serialize(value),
      'trace_id': self.__trace.id,
    }

    trace_alias = TraceAlias.find_by(**columns)
    if trace_alias:
      return trace_alias

    trace_alias = TraceAlias.create(**{
      **columns,
      'trace_request_id': trace_request.id if trace_request else None
    })

    if trace_alias:
      Logger.instance(LOG_ID).info(f"{bcolors.OKGREEN}Created{bcolors.ENDC} alias {trace_alias.name}: {value}")

    return trace_alias

  def __resolve_alias(self, alias_name: str, value: str):
    strategy = self.__strategy

    if strategy == alias_resolve_strategy.NONE:
      return

    trace_aliases = self.__resolve_aliases(alias_name, value)
    if trace_aliases.is_empty():
      return
      
    if strategy == alias_resolve_strategy.LIFO:
      return trace_aliases.pop()
    elif strategy == alias_resolve_strategy.FIFO:
      return trace_aliases.shift()

  def __resolve_aliases(self, alias_name: str, value: str) -> Collection:
    '''
    Return TraceAlias collection based on alias_name and value
    '''
    trace_aliases = self.__assigned_aliases(alias_name, value)

    if trace_aliases.is_empty():
      trace_aliases = self.__available_aliases(alias_name)

    return trace_aliases

  def __assigned_aliases(self, alias_name: str, value: str) -> Collection:
    trace = self.__trace

    trace_alias_hash = {
      'assigned_to': self.__serialize(value),
      'name': alias_name,
      'trace_id': trace.id,
    }

    trace_aliases = TraceAlias.where(trace_alias_hash).get()

    if not trace_aliases.is_empty():
      self.__log_resolved_aliases(trace_aliases)

    return trace_aliases

  def __available_aliases(self, alias_name: str) -> Collection:
    trace = self.__trace

    trace_aliases = TraceAlias.where({
      'name': alias_name,
      'trace_id': trace.id,
    }).where_null('assigned_to').get()

    if not trace_aliases.is_empty():
      self.__log_resolved_aliases(trace_aliases)

    return trace_aliases

  def __serialize(self, value):
    return str(value) if isinstance(value, list) or isinstance(value, dict) else value

  def __log_resolved_aliases(self, trace_aliases):
    trace_aliases.each(lambda trace_alias: Logger.instance(LOG_ID).debug(f"\tResolved Trace Alias: {trace_alias.to_dict()}"))


import importlib
import os
import pdb
import pytest

from click.testing import CliRunner

from stoobly_agent.app.cli import config_cli
from stoobly_agent.app.settings import Settings
from stoobly_agent.config.constants import env_vars
from stoobly_agent.test.test_helper import reset
from stoobly_agent.lib.api.keys import ProjectKey

@pytest.fixture(scope='module')
def config():
  importlib.reload(env_vars) # Fix strange bug where FEATURE_REMOTE does not exist
  os.environ[env_vars.FEATURE_REMOTE] = '1'
  importlib.reload(config_cli)
  del os.environ[env_vars.FEATURE_REMOTE]
  return config_cli.config

class TestConfigProjectSet():

  @pytest.fixture(scope='class', autouse=True)
  def settings(self):
    return reset()

  class TestWhenRecordPolicyOverwrite():

    def test_it_sets_scenario_overwritable(self, config, settings: Settings):
      settings.proxy.intercept.active = True
      settings.commit()

      project_key = settings.proxy.intercept.project_key
      _project_key = ProjectKey(project_key)

      set_results = CliRunner().invoke(config, ['project', 'set', _project_key.encode(1, _project_key.organization_id)])
      assert set_results.exit_code == 0

      settings.load()
      assert settings.proxy.intercept.active == False



from typing import TypedDict

from stoobly_agent.lib.api.interfaces.pagination_query_params import PaginationQueryParams

class ReplayedResponseIndexQueryParams(PaginationQueryParams):
  request_id: int

class ReplayeResponseCreateParams:
  latency: int
  raw: bytes
  request_id: int
  status: int

class ReplayedResponseShowParams:
  id: int
  latency: int
  raw: bytes
  request_id: int
  status: int


import click
import pdb

from stoobly_agent.app.cli.helpers.handle_replay_service import JSON_FORMAT
from stoobly_agent.app.cli.helpers.print_service import FORMATS
from stoobly_agent.app.models.factories.resource.local_db.helpers.log_event import DELETE_ACTION, PUT_ACTION
from stoobly_agent.app.settings import Settings
from stoobly_agent.config.constants import alias_resolve_strategy, test_filter, test_output_level, test_strategy
from stoobly_agent.lib import logger
from stoobly_agent.lib.utils.conditional_decorator import ConditionalDecorator

from .handlers.scenario_cli_handler import (
    create_handler, delete_handler, list_handler, replay_handler, reset_handler, show_handler, snapshot_handler, test_handler
)
from .helpers.feature_flags import local, remote
from .helpers.validations import *
from .types.scenario import ScenarioTestOptions

settings = Settings.instance()
is_remote = remote(settings)
is_local = local(settings)

@click.group(
    epilog="Run 'stoobly-agent scenario COMMAND --help' for more information on a command.",
    help="Manage request scenarios"
)
@click.pass_context
def scenario(ctx):
    pass

@scenario.command(
    help="Create a scenario"
)
@click.option('--description', default='', help='Scenario description.')
@click.option('--format', type=click.Choice(FORMATS), help='Format output.')
@ConditionalDecorator(lambda f: click.option('--project-key', help='Project to create scenario in.')(f), is_remote)
@click.option('--select', multiple=True, help='Select column(s) to display.')
@click.option('--without-headers', is_flag=True, default=False, help='Disable printing column headers.')
@click.argument('name')
def create(**kwargs):
    create_handler(kwargs)

@scenario.command(
    help="Replay a scenario"
)
@ConditionalDecorator(lambda f: click.option(
    '--alias-resolve-strategy', 
    default=alias_resolve_strategy.NONE, 
    type=click.Choice([alias_resolve_strategy.NONE, alias_resolve_strategy.FIFO, alias_resolve_strategy.LIFO]), 
    help='Strategy for resolving dynamic values for aliases.'
)(f), is_remote)
@ConditionalDecorator(lambda f: click.option('--assign', multiple=True, help='Assign alias values. Format: <NAME>=<VALUE>')(f), is_remote)
@click.option('--format', type=click.Choice([JSON_FORMAT]), help='Format replay response.')
@ConditionalDecorator(lambda f: click.option('--group-by', help='Repeat for each alias name.')(f), is_remote)
@click.option('--host', help='Rewrite request host.')
@click.option('--lifecycle-hooks-path', help='Path to lifecycle hooks script.')
@click.option(
    '--log-level', default=logger.WARNING, type=click.Choice([logger.DEBUG, logger.INFO, logger.WARNING, logger.ERROR]), 
    help=f'''
        Configure which logs to print. Defaults to {logger.WARNING}.
    '''
)
@ConditionalDecorator(
  lambda f: click.option(
    '--overwrite', is_flag=True, default=False, help='Replay scenario and overwrite existing responses.'
  )(f), not is_remote
)
@click.option('--record', is_flag=True, default=False, help='Replay scenario and record.')
@ConditionalDecorator(lambda f: click.option('--save', is_flag=True, default=False, help='Replay scenario and save to history.')(f), not is_remote)
@click.option('--scenario-key', help='Record to scenario.')
@click.option('--scheme', help='Rewrite request scheme.')
@ConditionalDecorator(lambda f: click.option('--trace-id', help='Use existing trace.')(f), is_remote)
@ConditionalDecorator(lambda f: click.option('--validate', multiple=True, help='Validate one or more aliases. Format: <NAME>=?<TYPE>')(f), is_remote)
@click.argument('key')
def replay(**kwargs):
   replay_handler(kwargs) 

@scenario.command(
    help="Show all scenarios"
)
@click.option('--format', type=click.Choice(FORMATS), help='Format output.')
@click.option('--page', default=0)
@ConditionalDecorator(lambda f: click.option('--project-key', help='Project to create scenario in.')(f), is_remote)
@click.option('--select', multiple=True, help='Select column(s) to display.')
@click.option('--search', help='Query to filter scenarios by')
@click.option('--size', default=10)
@click.option('--sort-by', default='created_at', help='created_at|name')
@click.option('--sort-order', default='desc', help='asc | desc')
@click.option('--without-headers', is_flag=True, default=False, help='Disable printing column headers.')
def list(**kwargs):
    list_handler(kwargs)    

@scenario.command(
    help="Describe scenario"
)
@click.option('--format', type=click.Choice(FORMATS), help='Format output.')
@click.option('--select', multiple=True, help='Select column(s) to display.')
@click.option('--without-headers', is_flag=True, default=False, help='Disable printing column headers.')
@click.argument('key')
def show(**kwargs):
    show_handler(kwargs)

@scenario.command(
  help="Delete a scenario"
)
@click.argument('scenario_key')
def delete(**kwargs):
    delete_handler(kwargs)

if is_local:
    @scenario.command(
        help="Snapshot a scenario"
    )
    @click.option('--action', default=PUT_ACTION, type=click.Choice([DELETE_ACTION, PUT_ACTION]), help='Sets snapshot action.')
    @click.option('--decode', default=False, is_flag=True, help="Toggles whether to decode response bodies.")
    @click.argument('scenario_key')
    def snapshot(**kwargs):
        snapshot_handler(kwargs)

    @scenario.command(
        help="Reset a scenario to its snapshot state"
    )
    @click.option('--force', default=False, is_flag=True, help="Toggles whether resources are hard deleted.")
    @click.argument('scenario_key')
    def reset(**kwargs):
        reset_handler(kwargs)

@scenario.command(
    help="Replay and test a scenario"
)
@click.option('--aggregate-failures', default=False, is_flag=True, help='Toggles whether to continue execution on failure.')
@click.option(
    '--alias-resolve-strategy', 
    default=alias_resolve_strategy.NONE, 
    type=click.Choice([alias_resolve_strategy.NONE, alias_resolve_strategy.FIFO, alias_resolve_strategy.LIFO]), 
    help='Strategy for resolving dynamic values for aliases.'
)
@click.option('--assign', multiple=True, help='Assign alias values. Format: <NAME>=<VALUE>')
@click.option(
    '--filter', 
    default=test_filter.ALL, 
    type=click.Choice([test_filter.ALL, test_filter.ALIAS, test_filter.LINK]), 
    help='For iterable responses, selectively test properties.'
)
@click.option('--format', type=click.Choice([JSON_FORMAT]), help='Format replay response.')
@click.option('--group-by', help='Repeat for each alias name.')
@click.option('--host', help='Rewrite request host.')
@click.option('--lifecycle-hooks-path', help='Path to lifecycle hooks script.')
@click.option(
    '--log-level', default=logger.WARNING, type=click.Choice([logger.DEBUG, logger.INFO, logger.WARNING, logger.ERROR]), 
    help=f'''
        Configure which logs to print. Defaults to {logger.WARNING}.
    '''
)
@click.option(
  '--output-level', default=test_output_level.PASSED, type=click.Choice([test_output_level.FAILED, test_output_level.SKIPPED, test_output_level.PASSED]),
  help=f'''
    Configure which tests to print. Defaults to {test_output_level.PASSED}.
  '''
)
@click.option('--public-directory-path', help='Path to public files. Used for mocking requests.')
@ConditionalDecorator(lambda f: click.option('--remote-project-key', help='Use remote project for endpoint definitions.')(f), is_remote)
@ConditionalDecorator(lambda f: click.option('--report-key', help='Save results to report.')(f), is_remote)
@click.option('--response-fixtures-path', help='Path to response fixtures yaml. Used for mocking requests.')
@ConditionalDecorator(lambda f: click.option('--save', is_flag=True, default=False, help='Save results.')(f), is_remote)
@click.option('--scheme', help='Rewrite request scheme.')
@click.option(
    '--strategy', 
    default=test_strategy.DIFF, 
    type=click.Choice([test_strategy.CONTRACT, test_strategy.CUSTOM, test_strategy.DIFF, test_strategy.FUZZY]), 
    help='How to test responses.'
)
@click.option('--trace-id', help='Use existing trace.')
@click.option('--validate', multiple=True, help='Validate one or more aliases. Format: <NAME>=?<TYPE>')
@click.argument('key')
def test(**kwargs: ScenarioTestOptions):
    test_handler(kwargs)


from mitmproxy.http import Response as MitmproxyResponse

from ..context_response import TestContextResponse

class MitmproxyResponseAdapter():

  def __init__(self, response: MitmproxyResponse) -> None:
    self._response = response

  def adapt(self):
    response = TestContextResponse()
    response.with_headers(self._response.headers)
    response.with_content(self._response.content)
    response.with_status_code(self._response.status_code)
    return response


from io import BytesIO
import requests

from mitmproxy.http import Response as MitmproxyResponse
from urllib3 import HTTPResponse

class PythonResponseAdapter():

  def __init__(self, response: MitmproxyResponse):
    self.__response = response

  def adapt(self) -> requests.Response:
    response = requests.Response()
    response.status_code = self.__response.status_code
    response.headers = self.__response.headers

    response.raw = HTTPResponse(
      body=BytesIO(self.__response.raw_content),
      decode_content=False,
      headers=self.__response.headers,
      preload_content=False
    )

    return response


import dns.resolver
import os
import pdb
import subprocess
import re
import yaml

from typing import TypedDict

from stoobly_agent.config.data_dir import DataDir
from stoobly_agent.lib.logger import Logger

from .app import App
from .constants import (
  APP_DIR_ENV, APP_NETWORK_ENV, CA_CERTS_DIR_ENV, CERTS_DIR_ENV, CONTEXT_DIR_ENV, NAMESERVERS_FILE, 
  SERVICE_DNS_ENV, SERVICE_NAME_ENV, SERVICE_SCRIPTS_DIR,  SERVICE_SCRIPTS_ENV, USER_ID_ENV, 
  WORKFLOW_NAME_ENV, WORKFLOW_NAMESPACE_ENV, WORKFLOW_SCRIPTS_DIR, WORKFLOW_SCRIPTS_ENV, WORKFLOW_TEMPLATE_ENV
)
from .docker.constants import APP_EGRESS_NETWORK_TEMPLATE, APP_INGRESS_NETWORK_TEMPLATE, DOCKERFILE_CONTEXT
from .workflow_command import WorkflowCommand
from .workflow_env import WorkflowEnv

LOG_ID = 'WorkflowRunCommand'

class ComposeOptions(TypedDict):
  namespace: str
  user_id: str

class BuildOptions(ComposeOptions):
  user_id: str
  verbose: bool

class DownOptions(ComposeOptions):
  extra_compose_path: str
  rmi: bool

class UpOptions(ComposeOptions):
  attached: bool
  extra_compose_path: str
  pull: bool

class WorkflowRunCommand(WorkflowCommand):
  def __init__(self, app: App, **kwargs):
    super().__init__(app, **kwargs)

    self.__app_dir_path = os.path.abspath(kwargs.get('app_dir_path') or app.dir_path)
    self.__current_working_dir = os.getcwd()
    self.__ca_certs_dir_path = kwargs.get('ca_certs_dir_path') or app.ca_certs_dir_path
    self.__certs_dir_path = kwargs.get('certs_dir_path') or app.certs_dir_path
    self.__context_dir_path = kwargs.get('context_dir_path') or app.context_dir_path
    self.__network = kwargs.get('network') or self.app_config.network

  @property
  def app_dir_path(self):
    return self.__app_dir_path

  @property
  def ca_certs_dir_path(self):
    return self.__ca_certs_dir_path

  @property
  def certs_dir_path(self):
    return self.__certs_dir_path

  @property
  def context_dir_path(self):
    if not self.__context_dir_path:
      data_dir = DataDir.instance()
      return os.path.dirname(data_dir.path) 

    return self.__context_dir_path

  @property
  def current_working_dir(self):
    return self.__current_working_dir

  @current_working_dir.setter
  def current_working_dir(self, v):
    self.__current_working_dir = v

  @property
  def nameservers(self):
    path = self.nameservers_path
    if not os.path.exists(path):
      return []

    with open(path, 'r') as fp:
      nameservers = fp.read()
      return nameservers.split("\n")

  @property
  def nameservers_path(self):
    return os.path.join(self.app.data_dir.tmp_dir_path, NAMESERVERS_FILE)

  @property
  def network(self):
    return self.__network

  def create_image(self, **options: BuildOptions):
    relative_namespace_path = os.path.relpath(self.scaffold_namespace_path, self.__current_working_dir)
    dockerfile_path = os.path.join(relative_namespace_path, DOCKERFILE_CONTEXT)
    user_id = options['user_id'] or os.getuid()
    
    command = ['docker', 'build']
    command.append(f"-f {dockerfile_path}")
    command.append(f"-t stoobly.{user_id}")
    command.append(f"--build-arg USER_ID={user_id}")

    if not os.environ.get('STOOBLY_IMAGE_USE_LOCAL'):
      command.append('--pull')

    if not options.get('verbose'):
      command.append('--quiet')

    # To avoid large context transfer times, should be a folder with relatively low number of files
    command.append(relative_namespace_path) 

    return ' '.join(command)

  def remove_image(self, user_id: str = None):
    user_id = user_id or os.getuid()
    command = ['docker', 'rmi', f"stoobly.{user_id}", '&>', '/dev/null']
    command.append('|| true')
    return ' '.join(command)

  def create_egress_network(self):
    return f"docker network create {APP_EGRESS_NETWORK_TEMPLATE.format(network=self.network)} &> /dev/null"

  def create_ingress_network(self):
    return f"docker network create {APP_INGRESS_NETWORK_TEMPLATE.format(network=self.network)} &> /dev/null"

  def remove_egress_network(self):
    return f"docker network rm {APP_EGRESS_NETWORK_TEMPLATE.format(network=self.network)} &> /dev/null || true"

  def remove_ingress_network(self):
    return f"docker network rm {APP_INGRESS_NETWORK_TEMPLATE.format(network=self.network)} &> /dev/null || true"

  def up(self, **options: UpOptions):
    if not os.path.exists(self.compose_path):
      return ''

    command = ['COMPOSE_IGNORE_ORPHANS=true', 'docker', 'compose']
    command_options = []

    # Add docker compose file
    command_options.append(f"-f {os.path.relpath(self.compose_path, self.__current_working_dir)}")

    # Add docker compose networks file
    command_options.append(f"-f {os.path.relpath(self.networks_compose_path, os.getcwd())}")

    # Add custom docker compose file
    custom_services = self.custom_services
    if custom_services:
      uses_profile = False
      for service_name in custom_services:
        service = custom_services[service_name]
        profiles = service.get('profiles')
        if isinstance(profiles, list):
          if self.workflow_name in profiles:
            uses_profile = True
            break
      if not uses_profile:
        # TODO: looking into why warning does not print in docker
        Logger.instance(LOG_ID).error(f"Missing {self.workflow_name} profile in custom compose file")

      compose_file_path = os.path.relpath(self.custom_compose_path, self.__current_working_dir)
      command_options.append(f"-f {compose_file_path}")

    if options.get('extra_compose_path'):
      compose_file_path = os.path.relpath(options['extra_compose_path'], self.__current_working_dir)
      command_options.append(f"-f {compose_file_path}")

    command_options.append(f"--profile {self.workflow_name}") 

    if not options.get('namespace'):
      options['namespace'] = self.workflow_name
    command_options.append(f"-p {options['namespace']}")

    command += command_options
    command.append('up')

    if not options.get('no_build'):
      command.append('--build')

    if options.get('pull'):
      command.append('--pull missing')

    if not options.get('attached'):
      command.append('-d')
    else:
      major_version = 2
      minor_version = 27
      patch_version = 0
      min_version = major_version * 10000 + minor_version * 100 + patch_version
      formula = "'{print $1*10000 + $2*100 + $3}'"
      option = f"$(test $(echo $(docker compose version --short) | awk -F. {formula}) -ge {min_version} && echo '--abort-on-container-failure')"

      # This option enables docker compose to return exit code 1 
      # when one of the services exits with a non-zero exit code
      # Otherwise, even if a service exits with a non-zero exit code, exit code 0 is returned
      command.append(option)

    self.write_env(**options)

    return ' '.join(command)

  def down(self, **options: DownOptions):
    if not os.path.exists(self.compose_path):
      return ''
  
    command = ['docker', 'compose']

    # Add docker compose file
    command.append(f"-f {os.path.relpath(self.compose_path, os.getcwd())}")

    # Add docker compose networks file
    command.append(f"-f {os.path.relpath(self.networks_compose_path, os.getcwd())}")

    # Add custom docker compose file
    if self.custom_services:
      command.append(f"-f {os.path.relpath(self.custom_compose_path, self.__current_working_dir)}")

    if options.get('extra_compose_path'):
      command.append(f"-f {os.path.relpath(options['extra_compose_path'], self.__current_working_dir)}")

    command.append(f"--profile {self.workflow_name}") 

    if not options.get('namespace'):
      options['namespace'] = self.workflow_name
    command.append(f"-p {options['namespace']}")

    command.append('down')
    command.append('--volumes')

    if options.get('rmi'):
      command.append('--rmi local')

    self.write_env(**options)

    return ' '.join(command)

  def write_nameservers(self):
    # If hostname is set then the service is external and we will need to configure the container's DNS.
    # If we don't configure the container's DNS, then Docker's embedded DNS will potentially
    # use configuration from the host's /etc/hosts file. The user may have configured their
    # /etc/hosts file to resolve requests to localhost
    #
    # See: 
    # https://forums.docker.com/t/docker-127-0-0-11-resolver-should-use-host-etc-hosts-file/55157
    # https://docs.docker.com/network/#dns-services
    # 
    # TODO: ideally we want to know if the service is built locally, if so, then no need to set DNS
    # since Docker's embedded DNS will resolve to it
    dns_resolver = dns.resolver.Resolver()

    with open(self.nameservers_path, 'w') as fp:
      nameservers = self.__find_nameservers(dns_resolver)
      ipv4_pattern = re.compile(r'^((25[0-5]|2[0-4][0-9]|1?[0-9][0-9]?)\.){3}(25[0-5]|2[0-4][0-9]|1?[0-9][0-9]?)$')
      nameservers = [ip for ip in nameservers if ipv4_pattern.match(ip)]
      if nameservers:
        fp.write("\n".join(nameservers))

  def write_env(self, **options: ComposeOptions):
    namespace = options.get('namespace')
    user_id = options.get('user_id')

    _config = {}
    _config[APP_DIR_ENV] = self.app_dir_path
    _config[CA_CERTS_DIR_ENV] = self.ca_certs_dir_path
    _config[CERTS_DIR_ENV] = self.certs_dir_path
    _config[CONTEXT_DIR_ENV] = self.context_dir_path
    _config[SERVICE_NAME_ENV] = self.service_name
    _config[SERVICE_SCRIPTS_ENV] = SERVICE_SCRIPTS_DIR
    _config[USER_ID_ENV] = user_id or os.getuid()
    _config[WORKFLOW_NAME_ENV] = self.workflow_name
    _config[WORKFLOW_SCRIPTS_ENV] = WORKFLOW_SCRIPTS_DIR
    _config[WORKFLOW_TEMPLATE_ENV] = self.workflow_name
    
    if namespace:
      _config[WORKFLOW_NAMESPACE_ENV] = namespace

    if self.network:
      _config[APP_NETWORK_ENV] = self.network

    # Specified DNS should prioritized, otherwise defaults to internal DNS
    nameservers = self.nameservers
    if nameservers:
      _config[SERVICE_DNS_ENV] = nameservers[0]
    else:
      _config[SERVICE_DNS_ENV] = '8.8.8.8'

    env_vars = self.config(_config)
    WorkflowEnv(self.workflow_path).write(env_vars)
    return env_vars

  def __find_nameservers(self, dns_resolver: dns.resolver.Resolver):
    nameservers = dns_resolver.nameservers

    # If systemd-resolved is not used
    if nameservers != ['127.0.0.53']:
      return nameservers

    # Run the `resolvectl status` command and capture its output
    result = subprocess.run(['resolvectl', 'status'], stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)

    # Check if the command ran successfully
    if result.returncode != 0:
      return []

    # Extract the DNS servers using a regular expression
    #dns_servers = re.findall(r'DNS Servers: ([\d.]+(?:, [\d.]+)*)', result.stdout)
    pattern = re.compile('DNS Servers:(.*?)DNS Domain', re.DOTALL)
    match = re.findall(pattern, result.stdout)

    if not match:
      return []
      
    # Split the DNS servers string into a list
    dns_servers = match[0].strip().split("\n")
    return list(map(lambda dns_server: dns_server.strip(), dns_servers))



import pdb
import requests

from stoobly_agent.app.settings import Settings
from stoobly_agent.lib.api.keys.project_key import LOCAL_PROJECT_ID, ProjectKey
from stoobly_agent.lib.logger import Logger

class Model():
  def __init__(self, settings: Settings, **options):
    self.adapter = None
    self.is_local = False
    self.settings = settings

    if not settings.cli.features.remote:
      self.as_local()
    else:
      if 'access_token' in options:
        if options.get('access_token'):
          self.as_remote()
        else:
          self.as_local() 
      else:
        project_key = ProjectKey(settings.proxy.intercept.project_key)

        if int(project_key.id) == LOCAL_PROJECT_ID:
          self.as_local()
        else:
          self.as_remote()

  # Override
  def as_local(self):
    pass

  # Override
  def as_remote(self):
    pass

  def handle_request_error(self, e: requests.exceptions.RequestException):
      response: requests.Response = e.response

      if response:
        Logger.instance().error(f"{response.status_code} {response.content}")
        return response.content, response.status_code
      else:
        return 'Unknown Error', 0


from typing import List

from .types.proxy_settings import ParameterRule as IParameterRule

class ParameterRule:

  def __init__(self, parameter_rule: IParameterRule):
    self.update(parameter_rule)

  @property 
  def modes(self):
    return self.__modes

  @property
  def name(self):
    return self.__name

  @property
  def value(self):
    return self.__value

  @property
  def type(self):
    return self.__type

  def update(self, parameter_rule: IParameterRule):
    self.__parameter_rule = parameter_rule
    self.__modes = self.__parameter_rule.get('modes')
    self.__name = self.__parameter_rule.get('name')
    self.__value = self.__parameter_rule.get('value')
    self.__type = self.__parameter_rule.get('type')

  def to_dict(self):
    return {
      'modes': self.__modes,
      'name': self.__name,
      'value': self.__value,
      'type': self.__type,
    }


from .resource_key import ResourceKey

class InvalidReportKey(Exception):
  pass

class ReportKey(ResourceKey):

  def __init__(self, key: str):
    super().__init__(key)

    if not self.id:
      raise InvalidReportKey('Missing id')

    if not self.project_id:
      raise InvalidReportKey('Missing project_id')

  @property
  def project_id(self) -> str:
    return self.get('p')

  @property
  def id(self) -> str:
    return self.get('i')


import importlib
import os
import pdb
import pytest

from click.testing import CliRunner
from mock import patch
from unittest.mock import MagicMock

from stoobly_agent.test.test_helper import DETERMINISTIC_GET_REQUEST_URL, reset

# Enable remote feature
from stoobly_agent.config.constants import env_vars
from stoobly_agent import cli
from stoobly_agent.lib.api.keys import ProjectKey
from stoobly_agent.lib.orm.request import Request
from stoobly_agent.lib.orm.response import Response

from stoobly_agent.app.proxy.intercept_settings import InterceptSettings
from stoobly_agent.app.models.factories.resource.local_db.request_adapter import LocalDBRequestAdapter

@pytest.fixture(scope='module')
def runner():
    return CliRunner()

@pytest.fixture(scope='module')
def mock():
    os.environ[env_vars.FEATURE_REMOTE] = '1'
    importlib.reload(cli)
    del [env_vars.FEATURE_REMOTE]
    return cli.mock

class TestCliMockIntegration():
    @pytest.fixture(scope='class', autouse=True)
    def settings(self):
        return reset()

    class TestWhenRemoteProjectKey():
        @pytest.fixture(scope='class')
        def project_key(self):
            return ProjectKey(ProjectKey.encode(1, 1))

        @pytest.fixture(scope='class', autouse=True)
        def recorded_request_one(self, runner: CliRunner):
            record_result = runner.invoke(cli.record, [DETERMINISTIC_GET_REQUEST_URL])
            assert record_result.exit_code == 0
            return Request.last()

        @pytest.fixture(scope='class')
        def search_endpoint(self, runner: CliRunner, mock, project_key: ProjectKey):
            with patch('stoobly_agent.app.proxy.mock.eval_request_service.inject_search_endpoint') as spy:
                mock_result = runner.invoke(mock, ['--remote-project-key', project_key.raw, DETERMINISTIC_GET_REQUEST_URL])
                assert mock_result.exit_code == 0

                return spy

        def test_it_calls_search_endpoint_once(self, search_endpoint: MagicMock):
            assert search_endpoint.call_count == 1

        def test_search_endpoint_has_intercept_settings_call_arg(self, search_endpoint: MagicMock):
            args = search_endpoint.call_args[0]
            assert len(args) == 1
            assert type(args[0]) is InterceptSettings

    class TestWhenNotFound():
        @pytest.fixture(scope='class')
        def project_key(self):
            return ProjectKey(ProjectKey.encode(1, 1))

        @pytest.fixture(scope='class')
        def spies(self, runner: CliRunner, mock, project_key: ProjectKey):
            @patch('stoobly_agent.app.proxy.mock.eval_request_service.inject_search_endpoint')
            @patch.object(
                LocalDBRequestAdapter,
                '_LocalDBRequestAdapter__handle_request_not_found',
                wraps=LocalDBRequestAdapter(Request, Response)._LocalDBRequestAdapter__handle_request_not_found # Mock, but preserve implementation
            )
            @patch.object(
                LocalDBRequestAdapter,
                '_LocalDBRequestAdapter__ignored_components', 
                wraps=LocalDBRequestAdapter(Request, Response)._LocalDBRequestAdapter__ignored_components
            )
            def spy_on(search_endpoint, handle_request_not_found, ignored_components: MagicMock):
                search_endpoint.return_value = lambda a, b, c: {}
                test_result = runner.invoke(mock, ['--remote-project-key', project_key.raw, DETERMINISTIC_GET_REQUEST_URL])
                assert test_result.exit_code == 1
                return [search_endpoint, handle_request_not_found, ignored_components]

            return spy_on()

        @pytest.fixture(scope='class')
        def search_endpoint(self, spies):
            return spies[2]

        @pytest.fixture(scope='class')
        def handle_request_not_found(self, spies):
            return spies[1]

        @pytest.fixture(scope='class')
        def ignored_components(self, spies):
            return spies[0]

        def test_it_calls_handle_request_not_found_spy_once(self, handle_request_not_found: MagicMock):
            assert handle_request_not_found.call_count == 1

        def test_it_calls_ignored_components_once(self, ignored_components: MagicMock):
            assert ignored_components.call_count == 1


import pdb

from stoobly_agent.app.models.factories.resource.local_db.helpers.log import Log
from stoobly_agent.app.models.factories.resource.local_db.helpers.request_snapshot import RequestSnapshot
from stoobly_agent.app.models.factories.resource.local_db.helpers.scenario_snapshot import ScenarioSnapshot
from stoobly_agent.app.proxy.record import REQUEST_STRING_CLRF, RequestStringControl
from stoobly_agent.app.settings import Settings
from stoobly_agent.lib.logger import bcolors

from ..request_model import RequestModel
from ..scenario_model import ScenarioModel
from .create_request_params_service import build_params

class Apply():

  def __init__(self, **options):
    self.__force = options.get('force') or False
    self.__logger = None
    self.__request_model = None
    self.__scenario_model = None

  @property
  def force(self):
    return self.__force

  @force.setter
  def force(self, v):
    self.__force = v

  @property
  def request_model(self):
    if not self.__request_model:
      self.__request_model = RequestModel(Settings.instance())
    return self.__request_model

  @property 
  def scenario_model(self):
    if not self.__scenario_model:
      self.__scenario_model = ScenarioModel(Settings.instance())
    return self.__scenario_model

  def with_logger(self, logger):
    self.__logger = logger
    return self

  def with_request_model(self, request_model: RequestModel):
    self.__request_model = request_model
    return self

  def with_scenario_model(self, scenario_model: ScenarioModel):
    self.__scenario_model = scenario_model
    return self

  def all(self):
    log = Log()

    unprocessed_events = log.unprocessed_events
    events_count = len(unprocessed_events) 

    if events_count == 0:
      return

    last_processed_event = None
    completed = True

    for event in unprocessed_events:
      if self.__logger:
        self.__logger(f"{bcolors.OKBLUE}Processing Event{bcolors.ENDC} {event.uuid}")

      results = event.apply(**self.__handlers())
      if results:
        status = results[1]
        if status == 0 or status >= 400:
          completed = False
          break

      last_processed_event = event

    if last_processed_event:
      last_unprocessed_event = unprocessed_events[events_count - 1]

      if last_processed_event.uuid == last_unprocessed_event.uuid:
        # All unproccessed events have been processed, we are up to date
        log.version = log.next_version()
      else:
        # Some failure occurred, update log to last processed event uuid
        log.version = log.next_version(last_processed_event.uuid)

    log.lock()

    return completed

  def request(self, uuid: str):
    result = self.__apply_put_request(uuid)
    if not result:
      return False

    status = result[1]
    return status == 200

  def scenario(self, uuid: str):
    result = self.__apply_put_scenario(uuid)
    if not result:
      return False

    status = result[1]
    return status == 200

  def single(self, uuid: str):
    log = Log()

    events = log.events
    events_count = len(events) 

    if events_count == 0:
      return False

    target_event = None
    for event in events:
      if event.uuid == uuid:
        target_event = event
        break

    if not target_event:
      return False

    if self.__logger:
      self.__logger(f"{bcolors.OKBLUE}Processing Event{bcolors.ENDC} {event.uuid}")

    results = event.apply(**self.__handlers())
    if results:
      status = results[1]
      if status == 0 or status >= 400:
        return False

    return True

  def __handlers(self):
    return {
      'handle_request_delete': self.__apply_delete_request,
      'handle_request_put': self.__apply_put_request,
      'handle_scenario_delete': self.__apply_delete_scenario,
      'handle_scenario_put': self.__apply_put_scenario,
    }

  def __apply_delete_request(self, uuid: str):
    res, status = self.request_model.destroy(uuid, force=self.__force)

    if status == 200:
      self.__logger(f"{bcolors.WARNING}Deleted Request{bcolors.ENDC} {uuid}") 
    else: 
      self.__logger(f"{bcolors.FAIL}{status}{bcolors.ENDC} {res}")

    return res, status

  def __apply_put_request(self, uuid: str):
    snapshot = RequestSnapshot(uuid)

    raw_request = snapshot.request
    if not raw_request:
      error = f"snapshot for request {uuid} not found" 
      self.__logger(f"{bcolors.WARNING}Skipping Request{bcolors.ENDC} {error}")
      return error, 301

    return self.__put_request(uuid, raw_request)

  def __apply_delete_scenario(self, uuid: str):
    res, status = self.scenario_model.destroy(uuid, force=self.__force)

    if self.__logger and status == 200:
      self.__logger(f"{bcolors.WARNING}Deleted Scenario{bcolors.ENDC} {uuid}")
    else: 
      self.__logger(f"{bcolors.FAIL}{status}{bcolors.ENDC} {res}")

    return res, status

  def __apply_put_scenario(self, uuid: str):
    snapshot = ScenarioSnapshot(uuid)
    metadata = snapshot.metadata

    if not metadata:
      error = f"snapshot for scenario {uuid} not found"
      self.__logger(f"{bcolors.WARNING}Skipping Scenario{bcolors.ENDC} {error}")
      return error, 301 

    res, status = self.scenario_model.show(uuid)
    if status == 404:
      res, status = self.scenario_model.create(**{
        **metadata,
        'uuid': uuid,
      })

      if self.__logger:
        if status == 200:
          self.__logger(f"{bcolors.OKGREEN}Created Scenario{bcolors.ENDC} {res['name']}") 
        else: 
          self.__logger(f"{bcolors.FAIL}{status}{bcolors.ENDC} {res}")
    else:
      res, status = self.scenario_model.update(uuid, **{
        **metadata,
        'is_deleted': False,
      })

      if self.__logger:
        if status == 200:
          self.__logger(f"{bcolors.OKCYAN}Updated Scenario{bcolors.ENDC} {res['name']}") 
        else:
          self.__logger(f"{bcolors.FAIL}{status}{bcolors.ENDC} {res}")

    if status != 200:
      return res, status
    else:
      return self.__apply_put_scenario_requests(res, snapshot)

  def __apply_put_scenario_requests(self, scenario, snapshot: ScenarioSnapshot = None):
    if not snapshot:
      snapshot = ScenarioSnapshot(scenario['uuid'])

    snapshot_requests = {}

    raw_requests = snapshot.requests
    for raw_request in raw_requests:
      toks = raw_request.split(REQUEST_STRING_CLRF.encode(), 1)

      if len(toks) != 2:
        return f"{snapshot.requests_path} contains an invalid request", 400

      control = RequestStringControl(toks[0])
      uuid = control.id
      res, status = self.__put_request(uuid, raw_request, scenario_id=scenario['id'])

      if status != 200:
        return res, status

      snapshot_requests[uuid] = res

    # Remove requests in scenario that don't exist in the snapshot
    scenario_requests = []
    res, status = self.request_model.index(scenario_id=scenario['id'])

    if status == 200:
      scenario_requests = res['list']

    for request in scenario_requests:
      if request['uuid'] in snapshot_requests:
        continue

      res, status = self.__apply_delete_request(request['uuid'])

    return res, status

  def __put_request(self, uuid: str, raw_request: bytes, **base_params):   
    res, status = self.request_model.show(uuid)

    if status == 404:
      request_params = build_params(raw_request)

      if not request_params:
          self.__logger(f"{bcolors.FAIL}{status}{bcolors.ENDC} failed to join request {uuid}")
          return res, status

      params = {
        **request_params,
        **base_params,
      }

      res, status = self.request_model.create(**params)

      if self.__logger and status == 200:
        self.__logger(f"{bcolors.OKGREEN}Created Request{bcolors.ENDC} {res['list'][0]['url']}")
      else: 
        self.__logger(f"{bcolors.FAIL}{status}{bcolors.ENDC} {res}")
    elif status == 200:
      params = {
        'is_deleted': False,
        'request': raw_request,
        **base_params,
      }
      res, status = self.request_model.update(uuid, **params)

      if self.__logger:
        if status == 200:
          self.__logger(f"{bcolors.OKCYAN}Updated Request{bcolors.ENDC} {res['url']}")
        else: 
          self.__logger(f"{bcolors.FAIL}{status}{bcolors.ENDC} {res}")

    return res, status


import pdb

from .builder import WorkflowBuilder

class BuildDecorator():

  def __init__(self, workflow_builder: WorkflowBuilder):
    self.__workflow_builder = workflow_builder

  @property
  def workflow_builder(self):
    return self.__workflow_builder

  def decorate(self):
    service_builder = self.__workflow_builder.service_builder

    services = self.workflow_builder.services
    app_name = self.workflow_builder.app
    app_service = services.get(app_name) or {}

    services[app_name] = { 
      **app_service,
      **{ 
          'extends': service_builder.build_extends_proxy_base(self.workflow_builder.dir_path),
          'profiles': self.workflow_builder.profiles,
        } 
    }